from __future__ import annotations as _annotations

from typing import TypeAlias

from acp.interfaces import Agent as _AcpAgent
from acp.schema import (
    AudioContentBlock,
    BlobResourceContents,
    EmbeddedResourceContentBlock,
    HttpMcpServer,
    ImageContentBlock,
    McpServerStdio,
    PlanEntry,
    ResourceContentBlock,
    SseMcpServer,
    TextContentBlock,
    TextResourceContents,
)

AcpAgent: TypeAlias = _AcpAgent
AgentPromptBlock: TypeAlias = (
    TextContentBlock
    | ImageContentBlock
    | AudioContentBlock
    | ResourceContentBlock
    | EmbeddedResourceContentBlock
)

__all__ = (
    "AcpAgent",
    "AgentPromptBlock",
    "AudioContentBlock",
    "BlobResourceContents",
    "EmbeddedResourceContentBlock",
    "HttpMcpServer",
    "ImageContentBlock",
    "McpServerStdio",
    "PlanEntry",
    "ResourceContentBlock",
    "SseMcpServer",
    "TextContentBlock",
    "TextResourceContents",
)



from __future__ import annotations as _annotations

import asyncio
import builtins
import importlib
from types import SimpleNamespace
from typing import Any, cast

import pytest
from acp.exceptions import RequestError
from acp.schema import EnvVariable, HttpHeader, PromptResponse
from pydantic_acp.runtime import _native_plan_runtime as native_plan_runtime_module
from pydantic_acp.runtime import _prompt_execution as prompt_execution_module
from pydantic_acp.runtime._agent_state import (
    assign_model,
    clear_selected_model_id,
    default_model,
    has_native_plan_tools,
    remember_default_model,
    selected_model_id,
    set_active_session,
    set_native_plan_tools_installed,
    set_selected_model_id,
    try_active_session,
)
from pydantic_acp.runtime.adapter import NativePlanGeneration, TaskPlan
from pydantic_acp.runtime.prompts import load_message_history
from pydantic_acp.runtime.session_surface import SessionSurface
from pydantic_acp.types import (
    HttpMcpServer,
    ImageContentBlock,
    McpServerStdio,
    PlanEntry,
    SseMcpServer,
)
from pydantic_ai import models as pydantic_models
from pydantic_ai.exceptions import UserError
from pydantic_ai.messages import (
    FunctionToolResultEvent,
    ModelResponse,
    PartDeltaEvent,
    PartStartEvent,
    RetryPromptPart,
    TextPart,
    ToolCallPart,
    ToolReturnPart,
)
from pydantic_ai.tools import DeferredToolRequests
from typing_extensions import Sentinel

from examples.pydantic.travel_agent import TravelPromptModelProvider

from .support import (
    UTC,
    AcpSessionContext,
    AdapterConfig,
    AdapterModel,
    Agent,
    ConfigOption,
    MemorySessionStore,
    ModelSelectionState,
    ModeState,
    OpenAICompactionBridge,
    Path,
    PrepareToolsBridge,
    PrepareToolsMode,
    RecordingClient,
    SessionConfigOptionBoolean,
    SessionInfoUpdate,
    SessionMode,
    TestModel,
    agent_message_texts,
    create_acp_agent,
    datetime,
    text_block,
)

_INVALID_TEST_VALUE = Sentinel("_INVALID_TEST_VALUE")


def _adapter(*, agent: Agent[Any, Any], config: AdapterConfig | None = None) -> Any:
    return create_acp_agent(
        agent=agent,
        config=config or AdapterConfig(session_store=MemorySessionStore()),
    )


def _stored_session(adapter: Any, session_id: str) -> Any:
    return adapter._config.session_store.get(session_id)


def _fake_run_result(output: Any) -> Any:
    return SimpleNamespace(
        output=output,
        new_messages=list,
        all_messages=list,
        all_messages_json=lambda: b"[]",
    )


class _MutableAgent:
    def __init__(self) -> None:
        self.model: str = "initial-model"
        self._acp_selected_model_id: str | int | None = None
        self._acp_active_session: AcpSessionContext | str | None = None
        self._acp_default_model: str | None = None
        self._acp_native_plan_tools_installed: bool = False


def test_session_model_selection_respects_provider_and_fallback_state(
    tmp_path: Path,
) -> None:
    agent = Agent(TestModel(custom_output_text="ok"))
    adapter = _adapter(agent=agent, config=AdapterConfig(session_store=MemorySessionStore()))
    response = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    session = _stored_session(adapter, response.session_id)

    cast("Any", agent).model = _INVALID_TEST_VALUE
    session.session_model_id = None
    model_state = asyncio.run(adapter._get_model_selection_state(session, agent))
    assert model_state is None

    session.session_model_id = "session-model"
    model_state = asyncio.run(adapter._get_model_selection_state(session, agent))
    assert model_state is not None
    assert model_state.current_model_id == "session-model"
    assert model_state.available_models[0].model_id == "session-model"

    selected_agent = Agent(TestModel(custom_output_text="selected"))
    set_selected_model_id(selected_agent, "selected-model")
    selected_adapter = _adapter(
        agent=selected_agent,
        config=AdapterConfig(
            session_store=MemorySessionStore(),
            available_models=[
                AdapterModel(
                    model_id="model-a",
                    name="Model A",
                    override=cast("Any", selected_agent.model),
                ),
            ],
        ),
    )
    session.session_model_id = None
    assert selected_adapter._resolve_current_model_id(session, selected_agent) == "selected-model"
    assert selected_adapter._resolve_model_id_from_value(selected_agent.model) == "model-a"
    assert selected_adapter._resolve_model_id_from_value("plain-model") == "plain-model"
    assert selected_adapter._resolve_model_id_from_value(cast("Any", _INVALID_TEST_VALUE)) is None

    class DemoModelsProvider:
        def set_model(self, session, agent, model_id):
            del session, agent, model_id

        def get_model_state(self, session, agent):
            del session, agent
            return ModelSelectionState(
                current_model_id=None,
                available_models=[],
            )

    provider_adapter = _adapter(
        agent=Agent(TestModel(custom_output_text="provider")),
        config=AdapterConfig(
            models_provider=DemoModelsProvider(),
            session_store=MemorySessionStore(),
        ),
    )
    provider_response = asyncio.run(provider_adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    provider_session = _stored_session(provider_adapter, provider_response.session_id)
    provider_session.config_values["model"] = "stale"
    set_response = asyncio.run(
        provider_adapter.set_session_model("provider-model", provider_response.session_id),
    )
    assert set_response is not None
    assert provider_session.session_model_id is None
    provider_state = asyncio.run(
        provider_adapter._get_model_selection_state(provider_session, agent),
    )
    assert provider_state is not None
    assert provider_state.current_model_id is None

    missing_response = asyncio.run(
        provider_adapter.set_session_model("missing", provider_response.session_id),
    )
    assert missing_response is not None
    assert "model" not in provider_session.config_values


def test_session_mode_and_config_updates_flow_through_providers(tmp_path: Path) -> None:
    class DemoModesProvider:
        def set_mode(self, session, agent, mode_id):
            del session, agent, mode_id

        def get_mode_state(self, session, agent):
            del session, agent
            return ModeState(
                current_mode_id="review",
                modes=[SessionMode(id="review", name="Review")],
            )

    class EmptyConfigProvider:
        def set_config_option(self, session, agent, config_id, value):
            del session, agent, config_id, value

        def get_config_options(self, session, agent):
            del session, agent

    class MatchingConfigProvider:
        def set_config_option(self, session, agent, config_id, value):
            del session, agent, config_id, value

        def get_config_options(self, session, agent):
            del session, agent
            return cast(
                "Any",
                [
                    SessionConfigOptionBoolean(
                        id="flag",
                        name="Flag",
                        type="boolean",
                        current_value=False,
                    ),
                ],
            )

    mode_agent = Agent(TestModel(custom_output_text="mode"))
    mode_adapter = _adapter(
        agent=mode_agent,
        config=AdapterConfig(
            modes_provider=DemoModesProvider(),
            session_store=MemorySessionStore(),
        ),
    )
    mode_response = asyncio.run(mode_adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    mode_session = _stored_session(mode_adapter, mode_response.session_id)
    mode_state = asyncio.run(mode_adapter.set_session_mode("review", mode_response.session_id))
    assert mode_state is not None
    assert mode_session.config_values["mode"] == "review"

    assert asyncio.run(mode_adapter.set_config_option("x", mode_response.session_id, True)) is None

    empty_adapter = _adapter(
        agent=Agent(TestModel()),
        config=AdapterConfig(
            config_options_provider=EmptyConfigProvider(),
            session_store=MemorySessionStore(),
        ),
    )
    empty_response = asyncio.run(empty_adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    assert (
        asyncio.run(empty_adapter.set_config_option("flag", empty_response.session_id, True))
        is None
    )

    matching_adapter = _adapter(
        agent=Agent(TestModel()),
        config=AdapterConfig(
            config_options_provider=MatchingConfigProvider(),
            session_store=MemorySessionStore(),
        ),
    )
    matching_response = asyncio.run(matching_adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    config_response = asyncio.run(
        matching_adapter.set_config_option("flag", matching_response.session_id, True),
    )
    assert config_response is not None


def test_native_plan_state_syncs_through_runtime_outputs(tmp_path: Path) -> None:
    def pass_through(ctx, tool_defs):
        del ctx  # pragma: no cover
        return list(tool_defs)  # pragma: no cover

    persisted_states: list[tuple[list[str], str | None]] = []

    class RecordingPlanPersistenceProvider:
        def persist_plan_state(self, session, agent, entries, plan_markdown):
            del session, agent
            persisted_states.append(([entry.content for entry in entries], plan_markdown))

    plan_bridge = PrepareToolsBridge(
        default_mode_id="plan",
        modes=[
            PrepareToolsMode(
                id="plan",
                name="Plan",
                prepare_func=pass_through,
                plan_mode=True,
            ),
        ],
    )
    agent = Agent(TestModel(custom_output_text="plan"), output_type=str)
    adapter = _adapter(
        agent=agent,
        config=AdapterConfig(
            capability_bridges=[plan_bridge],
            native_plan_persistence_provider=RecordingPlanPersistenceProvider(),
            session_store=MemorySessionStore(),
        ),
    )
    response = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    session = _stored_session(adapter, response.session_id)

    output_type = adapter._build_run_output_type(agent, session=session)
    assert output_type is not None
    assert output_type == [TaskPlan, DeferredToolRequests]
    assert adapter._contains_text_output([int, str]) is True
    assert adapter._contains_text_output([int, float]) is False
    assert adapter._contains_native_plan_generation([str, NativePlanGeneration]) is True

    native_output = NativePlanGeneration(
        plan_md="# Plan",
        plan_entries=[PlanEntry(content="Write the plan", priority="high", status="in_progress")],
    )
    assert (
        adapter._synchronize_native_plan_output(session, native_output, streamed_output=True) == ""
    )
    asyncio.run(adapter._persist_current_native_plan_state(session, agent=agent))
    assert session.plan_markdown == "# Plan"
    assert persisted_states == [(["Write the plan"], "# Plan")]
    assert adapter._format_native_plan(session) == (
        "# Plan\n\nCurrent plan entries:\n\n1. [in_progress] (high) Write the plan\n\n"
        "Use these 1-based entry numbers with `acp_update_plan_entry` and "
        "`acp_mark_plan_done`."
    )

    session.plan_markdown = None
    session.plan_entries = []
    assert adapter._format_native_plan(session) == "No plan has been recorded yet."

    session.plan_entries = [
        PlanEntry(content="Inspect", priority="medium", status="pending").model_dump(mode="json"),
    ]
    assert "Inspect" in adapter._format_native_plan(session)

    adapter._install_native_plan_tools(agent)
    tools = agent._function_toolset.tools
    get_plan_tool = cast("Any", tools["acp_get_plan"])
    set_plan_tool = cast("Any", tools["acp_set_plan"])
    assert get_plan_tool.function() == "No plan has been recorded yet."
    assert persisted_states[-1] == (["Write the plan"], "# Plan")
    assert get_plan_tool.prepare(None, get_plan_tool.function_schema) is not None
    assert set_plan_tool.prepare(None, set_plan_tool.function_schema) is None


def test_tool_based_plan_generation_keeps_agent_output_type_and_exposes_set_plan(
    tmp_path: Path,
) -> None:
    def pass_through(ctx, tool_defs):
        del ctx  # pragma: no cover
        return list(tool_defs)  # pragma: no cover

    plan_bridge = PrepareToolsBridge(
        default_mode_id="plan",
        modes=[
            PrepareToolsMode(
                id="plan",
                name="Plan",
                prepare_func=pass_through,
                plan_mode=True,
            ),
        ],
    )
    agent = Agent(TestModel(custom_output_text="plan"), output_type=str)
    adapter = _adapter(
        agent=agent,
        config=AdapterConfig(
            capability_bridges=[plan_bridge],
            session_store=MemorySessionStore(),
        ),
    )
    response = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    session = _stored_session(adapter, response.session_id)

    plan_bridge.set_config_option(
        session,
        agent,
        "plan_generation_type",
        "tools",
    )
    assert adapter._build_run_output_type(agent, session=session) == [
        str,
        DeferredToolRequests,
    ]

    adapter._install_native_plan_tools(agent)
    set_active_session(agent, session)
    tools = agent._function_toolset.tools
    set_plan_tool = cast("Any", tools["acp_set_plan"])
    assert set_plan_tool.prepare(None, set_plan_tool.function_schema) is not None


def test_native_plan_runtime_disables_progress_and_writes_without_plan_bridge(
    tmp_path: Path,
) -> None:
    adapter = _adapter(
        agent=Agent(TestModel(custom_output_text="plain")),
        config=AdapterConfig(session_store=MemorySessionStore()),
    )
    response = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    session = _stored_session(adapter, response.session_id)

    native_plan_runtime = adapter._prompt_runtime._native_plan_runtime
    assert native_plan_runtime.supports_native_plan_progress(session) is False
    assert native_plan_runtime.supports_native_plan_writes(session) is False


def test_adapter_runtime_mixins_delegate_prompt_and_session_helpers(
    tmp_path: Path,
) -> None:
    class DemoModelsProvider:
        def get_model_state(self, session, agent):
            del session, agent
            return ModelSelectionState(
                current_model_id="demo-model",
                available_models=[
                    AdapterModel(
                        model_id="demo-model",
                        name="Demo Model",
                        override=TestModel(custom_output_text="demo"),
                    ),
                ],
            )

        def set_model(self, session, agent, model_id):
            del session, agent
            return ModelSelectionState(
                current_model_id=model_id,
                available_models=[
                    AdapterModel(
                        model_id=model_id,
                        name="Configured Model",
                        override=TestModel(custom_output_text="configured"),
                    ),
                ],
            )

    class DemoModesProvider:
        def get_mode_state(self, session, agent):
            del session, agent
            return ModeState(
                current_mode_id="plan",
                modes=[SessionMode(id="plan", name="Plan")],
            )

        def set_mode(self, session, agent, mode_id):
            del session, agent
            return ModeState(
                current_mode_id=mode_id,
                modes=[SessionMode(id=mode_id, name=mode_id.title())],
            )

    class DemoConfigProvider:
        def get_config_options(self, session, agent) -> list[ConfigOption]:
            del session, agent
            return [
                SessionConfigOptionBoolean(
                    id="flag",
                    name="Flag",
                    type="boolean",
                    current_value=False,
                ),
            ]

        def set_config_option(
            self,
            session,
            agent,
            config_id,
            value,
        ) -> list[ConfigOption] | None:
            del session, agent
            if config_id != "flag" or not isinstance(value, bool):
                return None  # pragma: no cover
            return [
                SessionConfigOptionBoolean(
                    id="flag",
                    name="Flag",
                    type="boolean",
                    current_value=value,
                ),
            ]

    class PromptOverrideProvider:
        def get_prompt_model_override(
            self,
            session,
            agent,
            prompt,
            model_override,
        ):
            del session, agent, prompt, model_override

    agent = Agent(TestModel(custom_output_text="ok"))
    adapter = _adapter(
        agent=agent,
        config=AdapterConfig(
            models_provider=DemoModelsProvider(),
            modes_provider=DemoModesProvider(),
            config_options_provider=DemoConfigProvider(),
            prompt_model_override_provider=PromptOverrideProvider(),
            session_store=MemorySessionStore(),
        ),
    )
    response = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    session = _stored_session(adapter, response.session_id)

    assert (
        asyncio.run(
            adapter._resolve_prompt_model_override(
                session,
                agent,
                prompt=[text_block("hi")],
                model_override=None,
            ),
        )
        is None
    )
    assert asyncio.run(adapter._record_bridge_updates(session, agent)) is None

    model_state = asyncio.run(adapter._set_provider_model_state(session, agent, "override-model"))
    assert model_state is not None
    adapter._synchronize_session_model_selection(session, model_state)

    mode_state = asyncio.run(adapter._set_provider_mode_state(session, agent, "review"))
    assert mode_state is not None
    adapter._synchronize_mode_state(session, mode_state)

    assert asyncio.run(adapter._set_provider_config_options(session, agent, "flag", True)) is True
    assert (
        asyncio.run(
            adapter.resume_session(
                cwd=str(tmp_path),
                session_id=response.session_id,
                mcp_servers=[],
            ),
        ).modes
        is not None
    )


def test_agent_state_supports_legacy_attrs_and_runtime_storage(tmp_path: Path) -> None:
    session = AcpSessionContext(
        session_id="session-1",
        cwd=tmp_path,
        created_at=datetime.now(UTC),
        updated_at=datetime.now(UTC),
    )
    legacy_agent = _MutableAgent()
    legacy_agent._acp_active_session = "invalid-session"
    legacy_agent._acp_default_model = "legacy-default"
    legacy_agent._acp_selected_model_id = 123
    legacy_agent._acp_native_plan_tools_installed = True

    assert try_active_session(legacy_agent) is None
    assert default_model(legacy_agent) == "legacy-default"
    assert selected_model_id(legacy_agent) is None
    assert has_native_plan_tools(legacy_agent) is True

    set_selected_model_id(legacy_agent, "model-a")
    assert selected_model_id(legacy_agent) == "model-a"
    clear_selected_model_id(legacy_agent)
    assert legacy_agent._acp_selected_model_id is None

    plain_agent = _MutableAgent()
    del plain_agent._acp_active_session
    del plain_agent._acp_default_model
    del plain_agent._acp_native_plan_tools_installed
    del plain_agent._acp_selected_model_id
    assert try_active_session(plain_agent) is None
    assert default_model(plain_agent) is None
    assert selected_model_id(plain_agent) is None
    assert has_native_plan_tools(plain_agent) is False

    set_active_session(plain_agent, session)
    remember_default_model(plain_agent)
    set_native_plan_tools_installed(plain_agent)
    assign_model(plain_agent, "switched-model")

    assert try_active_session(plain_agent) is session
    assert default_model(plain_agent) == "initial-model"
    assert has_native_plan_tools(plain_agent) is True
    assert plain_agent.model == "switched-model"


def test_runtime_model_restore_and_error_paths_are_handled(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    agent = Agent(TestModel(custom_output_text="default"))
    adapter = _adapter(
        agent=agent,
        config=AdapterConfig(
            session_store=MemorySessionStore(),
            approval_bridge=cast("Any", _INVALID_TEST_VALUE),
        ),
    )
    response = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    session = _stored_session(adapter, response.session_id)

    assert adapter._restore_default_model(agent, session) is False
    assert adapter._restore_agent_default_model(agent) is False

    session.session_model_id = "model-a"
    assert adapter._restore_default_model(agent, session) is False

    adapter._remember_default_model(agent)
    default_model_id = adapter._resolve_model_id_from_value(agent.model)
    set_selected_model_id(agent, "selected")
    agent.model = "switched-model"
    assert adapter._restore_agent_default_model(agent) is True
    assert selected_model_id(agent) is None

    session.session_model_id = "switched-model"
    agent.model = "different"
    assert adapter._restore_default_model(agent, session) is True
    assert session.config_values["model"] == default_model_id

    session.session_model_id = None
    cast("Any", agent).model = _INVALID_TEST_VALUE
    assert asyncio.run(adapter._resolve_model_override(session, agent)) is None

    with pytest.raises(RequestError):
        adapter._resolve_runtime_model(agent, model_override=None)
    with pytest.raises(UserError):
        adapter._resolve_runtime_model(agent, model_override="not a model id\n")

    with pytest.raises(RequestError):
        asyncio.run(
            adapter._resolve_deferred_approvals(session=session, requests=cast("Any", None)),
        )

    assert asyncio.run(adapter._record_cancelled_approval(session, None)) is None
    assert (
        asyncio.run(
            adapter._handle_slash_command("unknown", argument=None, session=session, agent=agent),
        )
        is None
    )

    with pytest.raises(RequestError):
        adapter._resolve_selected_model(" ")
    with pytest.raises(RequestError):
        adapter._resolve_selected_model("codex:   ")

    real_import = builtins.__import__

    def fake_import(name, *args, **kwargs):
        if name == "codex_auth_helper":
            raise ImportError("missing")
        return real_import(name, *args, **kwargs)  # pragma: no cover

    monkeypatch.setattr(builtins, "__import__", fake_import)
    with pytest.raises(RequestError):
        adapter._resolve_selected_model("codex:gpt-5.4")


def test_sessions_normalize_cwd_and_serialize_mcp_servers(tmp_path: Path) -> None:
    agent = Agent(TestModel(custom_output_text="misc"))
    adapter = _adapter(
        agent=agent,
        config=AdapterConfig(
            available_models=[
                AdapterModel(
                    model_id="default:test",
                    name="Default",
                    override=cast("Any", agent.model),
                ),
            ],
            session_store=MemorySessionStore(),
        ),
    )
    response = asyncio.run(adapter.new_session(cwd="relative/path", mcp_servers=[]))
    session = _stored_session(adapter, response.session_id)

    assert session.cwd == Path.cwd() / "relative/path"
    adapter._update_session_mcp_servers(session, None)
    assert session.mcp_servers == []

    stdio_server = McpServerStdio(
        name="stdio",
        command="python",
        args=["server.py"],
        env=[EnvVariable(name="TOKEN", value="secret")],
    )
    http_server = HttpMcpServer(
        name="http",
        url="https://example.com",
        headers=[HttpHeader(name="Authorization", value="Bearer secret")],
        type="http",
    )
    sse_server = SseMcpServer(
        name="sse",
        url="https://example.com/sse",
        headers=[HttpHeader(name="X-Trace", value="trace-id")],
        type="sse",
    )

    assert adapter._serialize_mcp_server(stdio_server) == {
        "args": ["server.py"],
        "command": "python",
        "env": {"TOKEN": "secret"},
        "name": "stdio",
        "transport": "stdio",
    }
    assert adapter._serialize_mcp_server(http_server) == {
        "headers": {"Authorization": "Bearer secret"},
        "name": "http",
        "transport": "http",
        "url": "https://example.com",
    }
    assert adapter._serialize_mcp_server(sse_server) == {
        "headers": {"X-Trace": "trace-id"},
        "name": "sse",
        "transport": "sse",
        "url": "https://example.com/sse",
    }


def test_prompt_execution_handles_streaming_and_deferred_fallbacks(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    agent = Agent(TestModel(custom_output_text="unused"), output_type=str)
    adapter = _adapter(
        agent=agent,
        config=AdapterConfig(
            session_store=MemorySessionStore(),
            enable_generic_tool_projection=False,
        ),
    )
    response = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    session = _stored_session(adapter, response.session_id)

    assert asyncio.run(adapter._record_tool_updates(session, agent, [])) is None
    assert adapter._normalize_cwd("relative/path") == Path.cwd() / "relative/path"

    async def fake_stream_run(*args: Any, **kwargs: Any) -> tuple[Any, bool]:
        del args, kwargs
        return _fake_run_result("streamed"), True

    monkeypatch.setattr(adapter, "_should_stream_text_responses", lambda *args, **kwargs: True)
    monkeypatch.setattr(adapter, "_run_prompt_with_events", fake_stream_run)
    outcome = asyncio.run(
        adapter._run_prompt(
            agent=agent,
            prompt=[text_block("stream please")],
            session=session,
        ),
    )
    assert outcome.result.output == "streamed"
    assert outcome.streamed_output is True

    monkeypatch.setattr(adapter, "_should_stream_text_responses", lambda *args, **kwargs: False)

    async def fake_run(prompt_text: str | None, **kwargs: Any) -> Any:
        del prompt_text, kwargs
        return _fake_run_result("plain")

    cast("Any", agent).run = fake_run
    outcome = asyncio.run(
        adapter._run_prompt(
            agent=agent,
            prompt=[text_block("plain please")],
            session=session,
        ),
    )
    assert outcome.result.output == "plain"
    assert outcome.streamed_output is False

    calls: list[int] = []

    async def flaky_run(prompt_text: str | None, **kwargs: Any) -> Any:
        del prompt_text, kwargs
        calls.append(1)
        if len(calls) == 1:
            raise UserError("broken model")
        return _fake_run_result("recovered")

    cast("Any", agent).run = flaky_run
    monkeypatch.setattr(adapter, "_restore_default_model", lambda *args, **kwargs: True)
    outcome = asyncio.run(
        adapter._run_prompt(
            agent=agent,
            prompt=[text_block("recover please")],
            session=session,
        ),
    )
    assert outcome.result.output == "recovered"

    no_bridge_result = _fake_run_result(
        DeferredToolRequests(approvals=[ToolCallPart("approval-only", {"x": 1})]),
    )

    async def no_bridge_run(prompt_text: str | None, **kwargs: Any) -> Any:
        del prompt_text, kwargs
        return no_bridge_result

    cast("Any", agent).run = no_bridge_run
    monkeypatch.setattr(adapter, "_supports_deferred_approval_bridge", lambda: False)
    no_bridge_outcome = asyncio.run(
        adapter._run_prompt(
            agent=agent,
            prompt=[text_block("approval please")],
            session=session,
        ),
    )
    assert no_bridge_outcome.stop_reason == "end_turn"
    assert (
        adapter._prompt_runtime._execution._prepare_run_output_type(
            agent,
            session,
            str,
        )
        is str
    )

    bridge_adapter = _adapter(
        agent=agent,
        config=AdapterConfig(
            session_store=MemorySessionStore(),
            approval_bridge=cast("Any", _INVALID_TEST_VALUE),
            enable_generic_tool_projection=False,
        ),
    )
    bridge_response = asyncio.run(bridge_adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    bridge_session = _stored_session(bridge_adapter, bridge_response.session_id)
    assert (
        bridge_adapter._prompt_runtime._execution._prepare_run_output_type(
            agent,
            bridge_session,
            DeferredToolRequests,
        )
        is DeferredToolRequests
    )

    async def calls_present_run(prompt_text: str | None, **kwargs: Any) -> Any:
        del prompt_text, kwargs
        return _fake_run_result(
            DeferredToolRequests(
                calls=[ToolCallPart("needs-call", {"x": 1})],
                approvals=[ToolCallPart("needs-approval", {"y": 2})],
            ),
        )

    cast("Any", agent).run = calls_present_run
    bridge_outcome = asyncio.run(
        bridge_adapter._run_prompt(
            agent=agent,
            prompt=[text_block("calls please")],
            session=bridge_session,
        ),
    )
    assert bridge_outcome.stop_reason == "end_turn"

    async def approval_loop_run(prompt_text: str | None, **kwargs: Any) -> Any:
        del prompt_text, kwargs
        return _fake_run_result(
            DeferredToolRequests(approvals=[ToolCallPart("approval", {"z": 3})]),
        )

    cast("Any", agent).run = approval_loop_run

    async def unresolved_approvals(**kwargs: Any) -> Any:
        del kwargs
        return SimpleNamespace(
            cancelled=False,
            cancelled_tool_call=None,
            deferred_tool_results=None,
        )

    monkeypatch.setattr(
        bridge_adapter,
        "_resolve_deferred_approvals",
        unresolved_approvals,
    )
    with pytest.raises(RequestError):
        asyncio.run(
            bridge_adapter._run_prompt(
                agent=agent,
                prompt=[text_block("loop please")],
                session=bridge_session,
            ),
        )

    async def empty_stream(prompt_text: str | None, **kwargs: Any):
        del prompt_text, kwargs
        yield FunctionToolResultEvent(RetryPromptPart("retry", tool_name=None))
        yield FunctionToolResultEvent(ToolReturnPart("final_result", "done"))

    cast("Any", agent).run_stream_events = empty_stream
    with pytest.raises(RequestError):
        asyncio.run(
            type(adapter)._run_prompt_with_events(
                adapter,
                agent=agent,
                prompt_input="stream",
                run_kwargs={},
                session=session,
            ),
        )


def test_prompt_execution_compaction_paths_cover_empty_updates_and_retry_skips(
    tmp_path: Path,
) -> None:
    agent = Agent(TestModel(custom_output_text="unused"), output_type=str)
    adapter = _adapter(
        agent=agent,
        config=AdapterConfig(
            session_store=MemorySessionStore(),
            capability_bridges=[OpenAICompactionBridge(message_count_threshold=1)],
        ),
    )
    response = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    session = _stored_session(adapter, response.session_id)
    execution = adapter._prompt_runtime._execution

    assert execution._skip_compaction_providers() == frozenset({"openai"})

    recorded_updates: list[Any] = []

    async def record_update(_session: Any, update: Any) -> None:
        recorded_updates.append(update)  # pragma: no cover

    async def record_bridge_updates(_session: Any, _agent: Any) -> None:
        return None

    adapter._prompt_runtime._record_update = record_update
    execution.record_bridge_updates = record_bridge_updates

    assert asyncio.run(execution.record_tool_updates(session, agent, [])) is None
    assert recorded_updates == []

    adapter._should_stream_text_responses = lambda *args, **kwargs: True
    assert (
        asyncio.run(
            execution._record_execution_updates(
                session=session,
                agent=agent,
                result=_fake_run_result("streamed"),
                model_override=None,
                run_output_type=None,
            ),
        )
        is None
    )
    assert recorded_updates == []

    async def retry_only_stream(prompt_text: str | None, **kwargs: Any):
        del prompt_text, kwargs
        yield FunctionToolResultEvent(RetryPromptPart("retry", tool_name=None))

    cast("Any", agent).run_stream_events = retry_only_stream
    with pytest.raises(RequestError):
        asyncio.run(
            type(adapter)._run_prompt_with_events(
                adapter,
                agent=agent,
                prompt_input="stream",
                run_kwargs={},
                session=session,
            ),
        )


def test_prompt_execution_records_visible_compaction_updates(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    agent = Agent(TestModel(custom_output_text="unused"), output_type=str)
    adapter = _adapter(
        agent=agent,
        config=AdapterConfig(session_store=MemorySessionStore()),
    )
    response = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    session = _stored_session(adapter, response.session_id)
    execution = adapter._prompt_runtime._execution
    recorded_updates: list[Any] = []

    async def record_update(_session: Any, update: Any) -> None:
        recorded_updates.append(update)

    async def record_bridge_updates(_session: Any, _agent: Any) -> None:
        return None

    monkeypatch.setattr(
        prompt_execution_module,
        "build_compaction_updates",
        lambda *args, **kwargs: [
            SimpleNamespace(kind="compaction", args=args, kwargs=kwargs),
        ],
    )
    adapter._prompt_runtime._record_update = record_update
    execution.record_bridge_updates = record_bridge_updates

    asyncio.run(execution.record_tool_updates(session, agent, []))
    assert len(recorded_updates) == 1

    recorded_updates.clear()
    adapter._should_stream_text_responses = lambda *args, **kwargs: True
    asyncio.run(
        execution._record_execution_updates(
            session=session,
            agent=agent,
            result=_fake_run_result("streamed"),
            model_override=None,
            run_output_type=None,
        ),
    )
    assert len(recorded_updates) == 1


def test_cancel_stops_active_prompt_and_persists_terminal_history(
    tmp_path: Path,
) -> None:
    async def run_scenario() -> None:
        agent = Agent(TestModel(custom_output_text="unused"), output_type=str)
        adapter = _adapter(
            agent=agent,
            config=AdapterConfig(session_store=MemorySessionStore()),
        )
        client = RecordingClient()
        adapter.on_connect(client)

        response = await adapter.new_session(cwd=str(tmp_path), mcp_servers=[])
        started = asyncio.Event()
        cancelled = asyncio.Event()

        async def blocking_run_prompt(*, agent: Any, prompt: Any, session: Any) -> Any:
            del agent, prompt, session
            started.set()
            try:
                await asyncio.Event().wait()
            except asyncio.CancelledError:
                cancelled.set()
                raise
            raise AssertionError("prompt task should have been cancelled")  # pragma: no cover

        object.__setattr__(adapter, "_run_prompt", blocking_run_prompt)

        prompt_task = asyncio.create_task(
            adapter.prompt(
                prompt=[text_block("Long running task")],
                session_id=response.session_id,
                message_id="cancelled-user-message",
            ),
        )
        await started.wait()

        await adapter.cancel(response.session_id)
        prompt_response = await prompt_task

        assert cancelled.is_set() is True
        assert prompt_response.stop_reason == "cancelled"
        assert prompt_response.user_message_id == "cancelled-user-message"

        stored_session = _stored_session(adapter, response.session_id)
        assert stored_session.message_history_json is not None
        history = load_message_history(stored_session.message_history_json)
        assert isinstance(history[-1], ModelResponse)
        assert any(
            isinstance(part, TextPart) and "User stopped the run." in part.content
            for part in history[-1].parts
        )
        assert agent_message_texts(client)[-1].startswith("User stopped the run.")

    asyncio.run(run_scenario())


def test_prompt_emits_derived_title_without_session_metadata(tmp_path: Path) -> None:
    adapter = create_acp_agent(
        agent=Agent(TestModel(custom_output_text="ok")),
        config=AdapterConfig(session_store=MemorySessionStore()),
    )
    client = RecordingClient()
    adapter.on_connect(client)
    session = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))

    asyncio.run(
        adapter.prompt(
            prompt=[text_block("Explain TCP slow start")],
            session_id=session.session_id,
            message_id="title-message",
        ),
    )

    title_updates = [
        update for _, update in client.updates if isinstance(update, SessionInfoUpdate)
    ]
    assert title_updates
    assert title_updates[-1].title == "Explain TCP slow start"
    assert title_updates[-1].field_meta is None


def test_prompt_runtime_handles_edge_cases_without_corrupting_session_state(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    agent = Agent(TestModel(custom_output_text="unused"), output_type=str)
    adapter = _adapter(
        agent=agent,
        config=AdapterConfig(session_store=MemorySessionStore()),
    )
    response = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    session = _stored_session(adapter, response.session_id)

    async def failing_execute(**kwargs: Any) -> Any:
        del kwargs
        raise UserError("broken model")

    monkeypatch.setattr(adapter._prompt_runtime, "_execute_prompt", failing_execute)
    monkeypatch.setattr(adapter, "_restore_default_model", lambda *args, **kwargs: False)
    with pytest.raises(UserError, match="broken model"):
        asyncio.run(
            adapter._run_prompt(
                agent=agent,
                prompt=[text_block("recover")],
                session=session,
            ),
        )

    assert adapter._build_run_kwargs(
        session=session,
        message_history=None,
        deferred_tool_results=None,
        deps=None,
        model_override=None,
        model_settings=None,
        output_type=None,
    ) == {
        "deferred_tool_results": None,
        "conversation_id": session.session_id,
        "message_history": None,
        "metadata": {"pydantic_acp_session_id": session.session_id},
        "model": None,
    }
    adapter._config.approval_bridge = None
    assert adapter._build_run_output_type(agent, session=session) is None
    assert adapter._contains_text_output(NativePlanGeneration) is True
    assert adapter._contains_text_output(cast("Any", _INVALID_TEST_VALUE)) is False
    assert adapter._contains_native_plan_generation([int, NativePlanGeneration]) is True
    assert adapter._contains_native_plan_generation([int, float]) is False

    tool_start_event = PartStartEvent(index=0, part=ToolCallPart("echo", {"text": "hi"}))
    tool_delta_event = PartDeltaEvent(index=0, delta=cast("Any", _INVALID_TEST_VALUE))
    assert adapter._text_chunk_from_event(tool_start_event) is None
    assert adapter._text_chunk_from_event(tool_delta_event) is None
    assert (
        adapter._text_chunk_from_event(PartStartEvent(index=1, part=TextPart("hello"))) == "hello"
    )

    session.plan_entries = []
    session.plan_markdown = "# Saved Plan"
    assert adapter._format_native_plan(session) == "# Saved Plan"
    assert asyncio.run(adapter._prompt_runtime._emit_native_plan_update(session)) is None
    assert adapter._consume_native_plan_update(session) is False

    with pytest.raises(RequestError):
        adapter._prompt_runtime._replace_native_plan_entry(session, index=1, status="completed")

    session.plan_entries = [
        PlanEntry(content="Pending task", priority="low", status="pending").model_dump(mode="json"),
    ]
    with pytest.raises(RequestError):
        adapter._prompt_runtime._replace_native_plan_entry(session, index=2, status="completed")

    updated_entry = adapter._prompt_runtime._replace_native_plan_entry(
        session,
        index=1,
        content="Updated task",
        priority="high",
    )
    assert updated_entry.content == "Updated task"
    assert updated_entry.priority == "high"
    assert updated_entry.status == "pending"

    tool_agent = Agent(TestModel(custom_output_text="unused"), output_type=str)
    tool_adapter = _adapter(
        agent=tool_agent,
        config=AdapterConfig(
            approval_bridge=None,
            session_store=MemorySessionStore(),
        ),
    )
    tool_adapter._install_native_plan_tools(tool_agent)
    tools = tool_agent._function_toolset.tools
    get_plan_tool = cast("Any", tools["acp_get_plan"])
    set_plan_tool = cast("Any", tools["acp_set_plan"])
    update_plan_tool = cast("Any", tools["acp_update_plan_entry"])
    mark_done_tool = cast("Any", tools["acp_mark_plan_done"])

    session.config_values["mode"] = "ask"
    assert get_plan_tool.prepare(None, get_plan_tool.function_schema) is None
    assert get_plan_tool.function() == "No active ACP session is bound."
    assert set_plan_tool.prepare(None, set_plan_tool.function_schema) is None
    assert asyncio.run(update_plan_tool.function(index=1)) == "No active ACP session is bound."
    assert asyncio.run(mark_done_tool.function(index=1)) == "No active ACP session is bound."

    async def retry_and_output_stream(prompt_text: str | None, **kwargs: Any):
        del prompt_text, kwargs
        yield FunctionToolResultEvent(RetryPromptPart("retry", tool_name="output"))
        yield FunctionToolResultEvent(ToolReturnPart("final_result", "done"))

    cast("Any", agent).run_stream_events = retry_and_output_stream
    with pytest.raises(RequestError):
        asyncio.run(
            type(adapter)._run_prompt_with_events(
                adapter,
                agent=agent,
                prompt_input="stream",
                run_kwargs={},
                session=session,
            ),
        )


def test_native_plan_additional_instructions_append_without_replacing_core_guidance(
    tmp_path: Path,
) -> None:
    agent = Agent(TestModel(custom_output_text="wrapped"), output_type=str)
    adapter = _adapter(
        agent=agent,
        config=AdapterConfig(
            native_plan_additional_instructions=(
                "Keep plans concise.\nOnly use in-progress for multi-turn work."
            ),
            session_store=MemorySessionStore(),
        ),
    )
    response = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    session = _stored_session(adapter, response.session_id)

    session.plan_entries = [
        PlanEntry(content="Inspect repo", priority="medium", status="pending").model_dump(
            mode="json",
        ),
    ]

    rendered = adapter._format_native_plan(session)

    assert "Use these 1-based entry numbers with `acp_update_plan_entry`" in rendered
    assert "Additional plan instructions:" in rendered
    assert "Keep plans concise." in rendered
    assert "Only use in-progress for multi-turn work." in rendered
    assert "1. [pending] (medium) Inspect repo" in rendered


def test_native_plan_runtime_covers_empty_state_and_markdown_only_paths(
    tmp_path: Path,
) -> None:
    agent = Agent(TestModel(custom_output_text="wrapped"), output_type=str)
    adapter = _adapter(
        agent=agent,
        config=AdapterConfig(
            capability_bridges=[
                PrepareToolsBridge(
                    default_mode_id="plan",
                    modes=[
                        PrepareToolsMode(
                            id="plan",
                            name="Plan",
                            description="Structured plan mode.",
                            prepare_func=lambda ctx, tool_defs: list(tool_defs),
                            plan_mode=True,
                        ),
                    ],
                ),
            ],
            native_plan_additional_instructions="Keep the plan reviewable.",
            session_store=MemorySessionStore(),
        ),
    )
    client = RecordingClient()
    adapter.on_connect(client)
    response = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    session = _stored_session(adapter, response.session_id)
    native_plan_runtime = adapter._prompt_runtime._native_plan_runtime

    assert asyncio.run(native_plan_runtime.emit_native_plan_update(session)) is None
    assert (
        asyncio.run(native_plan_runtime.persist_current_native_plan_state(session, agent=agent))
        is None
    )

    session.plan_markdown = "# Saved plan\n"
    assert native_plan_runtime.format_native_plan(session) == (
        "# Saved plan\n\nAdditional plan instructions:\n\nKeep the plan reviewable."
    )

    adapter._config.native_plan_additional_instructions = "   "
    assert native_plan_runtime.format_native_plan(session) == "# Saved plan\n"


def test_native_plan_tool_prepare_paths_cover_transient_session_loss(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    tool_agent = Agent(TestModel(custom_output_text="unused"), output_type=str)
    tool_adapter = _adapter(
        agent=tool_agent,
        config=AdapterConfig(
            capability_bridges=[
                PrepareToolsBridge(
                    default_mode_id="agent",
                    modes=[
                        PrepareToolsMode(
                            id="agent",
                            name="Agent",
                            description="Execution mode with plan progress tools.",
                            prepare_func=lambda ctx, tool_defs: list(tool_defs),
                            plan_tools=True,
                        ),
                    ],
                ),
            ],
            session_store=MemorySessionStore(),
        ),
    )
    response = asyncio.run(tool_adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    session = _stored_session(tool_adapter, response.session_id)
    tool_adapter._install_native_plan_tools(tool_agent)
    tools = tool_agent._function_toolset.tools
    set_plan_tool = cast("Any", tools["acp_set_plan"])
    update_plan_tool = cast("Any", tools["acp_update_plan_entry"])

    calls = {"count": 0}

    def flaky_try_active_session(_agent: Any) -> Any:
        calls["count"] += 1
        return session if calls["count"] == 1 else None

    monkeypatch.setattr(native_plan_runtime_module, "try_active_session", flaky_try_active_session)

    assert update_plan_tool.prepare(None, update_plan_tool.function_schema) is None
    calls["count"] = 0
    assert set_plan_tool.prepare(None, set_plan_tool.function_schema) is None

    monkeypatch.setattr(native_plan_runtime_module, "try_active_session", lambda _agent: None)
    assert asyncio.run(set_plan_tool.function(entries=[])) == "No active ACP session is bound."


def test_prompt_model_override_provider_can_switch_model_for_media_prompts(
    tmp_path: Path,
) -> None:
    class DemoPromptModelOverrideProvider:
        def get_prompt_model_override(self, session, agent, prompt, model_override):
            del session, agent
            if any(isinstance(block, ImageContentBlock) for block in prompt):
                return "google-gla:gemini-3-flash-preview"
            return model_override  # pragma: no cover

    agent = Agent(TestModel(custom_output_text="ok"))
    adapter = _adapter(
        agent=agent,
        config=AdapterConfig(
            prompt_model_override_provider=DemoPromptModelOverrideProvider(),
            session_store=MemorySessionStore(),
        ),
    )
    response = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    session = _stored_session(adapter, response.session_id)

    resolved_override = asyncio.run(
        adapter._resolve_prompt_model_override(
            session,
            agent,
            prompt=[
                text_block("explain this image"),
                ImageContentBlock(type="image", data="aGVsbG8=", mime_type="image/png"),
            ],
            model_override="openrouter:google/gemini-3-flash-preview",
        ),
    )

    assert resolved_override == "google-gla:gemini-3-flash-preview"


def test_workspace_prompt_model_provider_prefers_explicit_media_override_for_image_prompts(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.setenv("ACP_TRAVEL_MEDIA_MODEL", "openai:gpt-4.1-mini")
    monkeypatch.delenv("TRAVEL_MEDIA_MODEL", raising=False)
    monkeypatch.setenv("MODEL_NAME", "openrouter:google/gemini-3-flash-preview")

    provider = TravelPromptModelProvider()
    session = cast("Any", object())
    agent = cast("Any", object())

    resolved_override = provider.get_prompt_model_override(
        session,
        agent,
        prompt=[
            text_block("describe"),
            ImageContentBlock(type="image", data="aGVsbG8=", mime_type="image/png"),
        ],
        model_override="openrouter:google/gemini-3-flash-preview",
    )

    assert resolved_override == "openai:gpt-4.1-mini"


def test_adapter_wrapper_methods_delegate_to_runtime_components(tmp_path: Path) -> None:
    agent = Agent(TestModel(custom_output_text="wrapped"), output_type=str)
    adapter = _adapter(agent=agent, config=AdapterConfig(session_store=MemorySessionStore()))
    response = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    session = _stored_session(adapter, response.session_id)

    active_calls: list[tuple[Any, Any]] = []
    apply_calls: list[tuple[Any, Any]] = []
    set_model_calls: list[tuple[Any, Any, Any]] = []

    def fake_set_active_session(a: Any, s: Any) -> None:
        active_calls.append((a, s))

    def fake_bind_session_client(s: Any) -> Any:
        return s

    async def fake_build_config_options(*args: Any, **kwargs: Any) -> list[str]:
        del args, kwargs
        return ["cfg"]

    async def fake_get_mode_state(*args: Any, **kwargs: Any) -> str:
        del args, kwargs
        return "mode"

    def fake_synchronize_mode_state(*args: Any, **kwargs: Any) -> None:
        del args, kwargs  # pragma: no cover

    async def fake_get_provider_config_options(*args: Any, **kwargs: Any) -> list[str]:
        del args, kwargs
        return ["opt"]

    async def fake_get_plan_entries(*args: Any, **kwargs: Any) -> list[str]:
        del args, kwargs
        return ["plan"]

    async def fake_synchronize_session_metadata(*args: Any, **kwargs: Any) -> None:
        del args, kwargs

    async def fake_get_approval_state(*args: Any, **kwargs: Any) -> dict[str, bool]:
        del args, kwargs
        return {"ok": True}

    def fake_find_model_option(*args: Any, **kwargs: Any) -> str:
        del args, kwargs
        return "model-option"

    def fake_require_model_option(*args: Any, **kwargs: Any) -> str:
        del args, kwargs
        return "required-model"

    def fake_supports_fallback_model_selection() -> bool:
        return True

    def fake_model_identity(value: Any) -> str:
        return f"id:{value}"

    def fake_apply_session_model_to_agent(a: Any, s: Any) -> None:
        apply_calls.append((a, s))

    def fake_set_agent_model(a: Any, s: Any, model_id: Any) -> None:
        set_model_calls.append((a, s, model_id))

    def fake_known_tool_call_starts(s: Any) -> dict[str, str]:
        del s
        return {"call": "start"}

    def fake_supports_streaming_model(*args: Any, **kwargs: Any) -> bool:
        del args, kwargs
        return True

    def fake_native_plan_bridge(s: Any) -> str:
        del s
        return "bridge"

    def fake_supports_native_plan_state(s: Any) -> bool:
        del s
        return True

    adapter._session_runtime._set_active_session = fake_set_active_session
    adapter._session_runtime._bind_session_client = fake_bind_session_client
    adapter._session_runtime._build_config_options = fake_build_config_options
    adapter._session_runtime._get_mode_state = fake_get_mode_state
    adapter._session_runtime._synchronize_mode_state = fake_synchronize_mode_state
    adapter._session_runtime._get_provider_config_options = fake_get_provider_config_options
    adapter._session_runtime._get_plan_entries = fake_get_plan_entries
    adapter._session_runtime._synchronize_session_metadata = fake_synchronize_session_metadata
    adapter._session_runtime._get_approval_state = fake_get_approval_state
    adapter._session_runtime._find_model_option = fake_find_model_option
    adapter._session_runtime._require_model_option = fake_require_model_option
    adapter._session_runtime._supports_fallback_model_selection = (
        fake_supports_fallback_model_selection
    )
    adapter._session_runtime._model_identity = fake_model_identity
    adapter._session_runtime._apply_session_model_to_agent = fake_apply_session_model_to_agent
    adapter._session_runtime._set_agent_model = fake_set_agent_model

    adapter._prompt_runtime._known_tool_call_starts = fake_known_tool_call_starts
    adapter._prompt_runtime._supports_streaming_model = fake_supports_streaming_model
    adapter._prompt_runtime._native_plan_bridge = fake_native_plan_bridge
    adapter._prompt_runtime._supports_native_plan_state = fake_supports_native_plan_state

    assert adapter._known_tool_call_starts(session) == {"call": "start"}
    adapter._set_active_session(agent, session)
    assert active_calls == [(agent, session)]
    assert adapter._bind_session_client(session) is session
    assert asyncio.run(
        adapter._build_config_options(
            session,
            agent,
            model_selection_state=None,
            mode_state=None,
        ),
    ) == ["cfg"]
    assert asyncio.run(adapter._get_mode_state(session, agent)) == "mode"
    assert asyncio.run(adapter._get_provider_config_options(session, agent)) == ["opt"]
    assert asyncio.run(adapter._get_plan_entries(session, agent)) == ["plan"]
    assert asyncio.run(adapter._synchronize_session_metadata(session, agent)) is None
    assert asyncio.run(adapter._get_approval_state(session, agent)) == {"ok": True}
    assert adapter._find_model_option("demo") == "model-option"
    assert adapter._require_model_option("demo") == "required-model"
    assert adapter._supports_fallback_model_selection() is True
    assert adapter._supports_streaming_model(agent, model_override=None) is True
    assert adapter._native_plan_bridge(session) == "bridge"
    assert adapter._supports_native_plan_state(session) is True
    assert adapter._model_identity("demo") == "id:demo"
    adapter._apply_session_model_to_agent(agent, session)
    adapter._set_agent_model(agent, session, "model-a")
    assert apply_calls == [(agent, session)]
    assert set_model_calls == [(agent, session, "model-a")]


def test_prompt_runtime_and_session_surface_cover_remaining_helper_edges(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    agent = Agent(TestModel(custom_output_text="helpers"), output_type=str)
    adapter = _adapter(agent=agent, config=AdapterConfig(session_store=MemorySessionStore()))
    response = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    session = _stored_session(adapter, response.session_id)

    native_plan_calls: list[tuple[str, Any]] = []

    def fake_supports_native_plan_progress(s: Any) -> bool:
        native_plan_calls.append(("supports_progress", s))
        return True

    async def fake_persist_external_plan_state(
        session: Any,
        *,
        agent: Any,
        entries: Any,
        plan_markdown: Any,
    ) -> None:
        native_plan_calls.append(
            ("persist_external", (session, agent, list(entries), plan_markdown)),
        )

    async def fake_persist_native_plan_state(
        session: Any,
        *,
        agent: Any,
        entries: Any,
        plan_markdown: Any,
    ) -> None:
        native_plan_calls.append(("persist_native", (session, agent, list(entries), plan_markdown)))

    adapter._prompt_runtime._native_plan_runtime.supports_native_plan_progress = (
        fake_supports_native_plan_progress
    )
    adapter._prompt_runtime._native_plan_runtime.persist_external_plan_state = (
        fake_persist_external_plan_state
    )
    adapter._prompt_runtime._native_plan_runtime.persist_native_plan_state = (
        fake_persist_native_plan_state
    )

    entries = [PlanEntry(content="One", priority="medium", status="pending")]
    assert adapter._prompt_runtime._known_tool_call_starts(session) == {}
    assert adapter._prompt_runtime._supports_native_plan_state(session) is False
    assert adapter._prompt_runtime._requires_native_plan_output(session) is False
    assert adapter._prompt_runtime._supports_native_plan_progress(session) is True
    assert adapter._prompt_runtime._get_native_plan_entries(session) is None
    asyncio.run(
        adapter._prompt_runtime._persist_external_plan_state(
            session,
            agent=agent,
            entries=entries,
            plan_markdown="# External",
        ),
    )
    asyncio.run(
        adapter._prompt_runtime._persist_native_plan_state(
            session,
            agent=agent,
            entries=entries,
            plan_markdown="# Native",
        ),
    )
    assert [call[0] for call in native_plan_calls] == [
        "supports_progress",
        "persist_external",
        "persist_native",
    ]

    prompt_model_runtime = adapter._prompt_runtime._model_runtime
    deferred_agent = cast("Any", SimpleNamespace(output_type=DeferredToolRequests))
    assert (
        prompt_model_runtime.build_run_output_type(deferred_agent, session=session)
        is DeferredToolRequests
    )

    hook_context = prompt_model_runtime.hook_context(agent=agent, session=session)
    with hook_context:
        pass

    no_hook_agent = Agent(TestModel(custom_output_text="helpers"), output_type=str)
    no_hook_adapter = _adapter(
        agent=no_hook_agent,
        config=AdapterConfig(
            session_store=MemorySessionStore(),
            hook_projection_map=None,
        ),
    )
    no_hook_response = asyncio.run(no_hook_adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    no_hook_session = _stored_session(no_hook_adapter, no_hook_response.session_id)
    with no_hook_adapter._prompt_runtime._model_runtime.hook_context(
        agent=no_hook_agent,
        session=no_hook_session,
    ):
        pass

    def fail_infer_model(value: Any) -> Any:
        raise ValueError(f"invalid model: {value}")

    monkeypatch.setattr(pydantic_models, "infer_model", fail_infer_model)
    with pytest.raises(UserError, match="invalid model"):
        prompt_model_runtime.resolve_runtime_model(agent, model_override="broken:model")

    assert prompt_model_runtime.contains_native_plan_generation([str, TaskPlan]) is True
    assert prompt_model_runtime.contains_native_plan_generation("text") is False

    session_model_runtime = adapter._session_runtime._model_runtime
    session_surface_runtime = adapter._session_runtime._surface_runtime

    async def unavailable_mode(mode_id: str, session_id: str) -> None:
        del mode_id, session_id

    async def unavailable_config(config_id: str, session_id: str, value: Any) -> None:
        del config_id, session_id, value

    adapter._session_runtime.set_session_mode = unavailable_mode
    adapter._session_runtime.set_config_option = unavailable_config

    async def fake_mode_state(*args: Any, **kwargs: Any) -> ModeState:
        del args, kwargs
        return ModeState(modes=[SessionMode(id="plan", name="Plan")], current_mode_id="plan")

    adapter._session_runtime._get_mode_state = fake_mode_state
    assert (
        asyncio.run(
            session_model_runtime.handle_slash_command(
                "plan",
                argument=None,
                session=session,
                agent=agent,
            ),
        )
        == "Mode `plan` is unavailable"
    )
    assert (
        asyncio.run(
            session_model_runtime.handle_slash_command(
                "thinking",
                argument="high",
                session=session,
                agent=agent,
            ),
        )
        == "Thinking effort is unavailable or invalid"
    )

    adapter._remember_default_model(agent)
    assign_model(agent, "restored-model")
    session.session_model_id = "restored-model"
    session.config_values["model"] = "restored-model"
    assert adapter._restore_default_model(agent, session) is True
    assert session.session_model_id == "test"
    assert session.config_values["model"] == "test"

    monkeypatch.setattr(
        adapter._session_runtime,
        "_resolve_model_id_from_value",
        lambda value: None,
    )
    assign_model(agent, "restored-model")
    session.session_model_id = "restored-model"
    session.config_values["model"] = "restored-model"
    assert adapter._restore_default_model(agent, session) is True
    assert session.session_model_id is None
    assert "model" not in session.config_values

    assert (
        asyncio.run(session_surface_runtime.set_provider_model_state(session, agent, "x")) is None
    )
    assert (
        asyncio.run(
            session_surface_runtime.set_provider_config_options(
                session,
                agent,
                "flag",
                True,
            ),
        )
        is False
    )
    assert session_surface_runtime.current_thinking_value(session, agent) is None

    monkeypatch.setattr(
        type(adapter._bridge_manager),
        "get_config_options",
        lambda self, s, a: [
            SessionConfigOptionBoolean(
                id="thinking",
                name="Thinking",
                current_value=True,
                type="boolean",
            ),
        ],
    )
    assert session_surface_runtime.current_thinking_value(session, agent) is None
    assert session_surface_runtime.plan_storage_metadata(session) is None

    with pytest.raises(ValueError):
        session_surface_runtime.require_model_option("missing-model")

    adapter._config.available_models = [
        AdapterModel(model_id="raw-model", name="Raw Model", override=cast("Any", agent.model)),
    ]
    assert session_surface_runtime.require_model_option("raw-model").model_id == "raw-model"

    assert session_surface_runtime.resolve_model_id_from_value("raw-model") == "raw-model"
    monkeypatch.setattr(adapter._session_runtime, "_model_identity", lambda value: None)
    assert session_surface_runtime.resolve_model_id_from_value("raw-model") == "raw-model"

    client = RecordingClient()
    adapter.on_connect(client)
    asyncio.run(
        session_surface_runtime.emit_session_state_updates(
            session,
            SessionSurface(
                config_options=None,
                model_state=None,
                mode_state=None,
                plan_entries=None,
            ),
            emit_available_commands=False,
            emit_config_options=False,
            emit_current_mode=True,
            emit_plan=False,
            emit_session_info=False,
        ),
    )
    assert client.updates == []

    async def no_model_selection_state(*args: Any, **kwargs: Any) -> None:
        del args, kwargs

    monkeypatch.setattr(
        session_surface_runtime,
        "get_model_selection_state",
        no_model_selection_state,
    )
    session.session_model_id = "unconfigured-model"
    clear_selected_model_id(agent)
    assert (
        asyncio.run(session_surface_runtime.resolve_model_override(session, agent))
        == "unconfigured-model"
    )
    assert adapter._prompt_runtime._supports_streaming_model(agent, model_override=None) is True
    adapter._set_native_plan_state(
        session,
        entries=[PlanEntry(content="Updated", priority="high", status="pending")],
        plan_markdown="# Updated",
    )
    assert session.plan_markdown == "# Updated"


def test_adapter_prompt_handler_covers_prompt_response_and_no_current_task(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    adapter = create_acp_agent(
        agent=Agent(TestModel(custom_output_text="ok")),
        config=AdapterConfig(session_store=MemorySessionStore()),
    )
    session_response = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    handler = cast("Any", adapter)._adapter_prompt

    adapter_prompt_module = importlib.import_module("pydantic_acp.runtime._adapter_prompt")
    monkeypatch.setattr(adapter_prompt_module.asyncio, "current_task", lambda: None)

    async def fake_run_prompt(**kwargs: Any) -> PromptResponse:
        del kwargs
        return PromptResponse(stop_reason="end_turn", usage=None, user_message_id="ignored")

    handler._run_prompt = fake_run_prompt
    response = asyncio.run(
        handler.prompt(
            [text_block("hello")],
            session_response.session_id,
            message_id="prompt-msg",
        ),
    )
    assert response.user_message_id == "prompt-msg"
    assert cast("Any", adapter)._active_prompt_tasks == {}

    cancelled = asyncio.run(
        handler._handle_cancelled_prompt(
            session=_stored_session(adapter, session_response.session_id),
            prompt_text="cancel me",
        ),
    )
    assert cancelled.stop_reason == "cancelled"


def test_unknown_slash_command_falls_through_to_prompt_execution(
    tmp_path: Path,
) -> None:
    adapter = create_acp_agent(
        agent=Agent(TestModel(custom_output_text="continued after slash")),
        config=AdapterConfig(session_store=MemorySessionStore()),
    )
    client = RecordingClient()
    adapter.on_connect(client)

    session = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    asyncio.run(
        adapter.prompt(
            prompt=[text_block("/unknown keep going")],
            session_id=session.session_id,
        ),
    )

    assert agent_message_texts(client) == ["continued after slash"]


def test_adapter_public_setters_cover_none_provider_and_mode_paths(
    tmp_path: Path,
) -> None:
    class EmptyModelsProvider:
        def set_model(self, session, agent, model_id):
            del session, agent, model_id

        def get_model_state(self, session, agent):
            del session, agent

    class ReviewModesProvider:
        def get_mode_state(self, session, agent):
            del session, agent
            return ModeState(
                current_mode_id="review",
                modes=[SessionMode(id="review", name="Review")],
            )

        def set_mode(self, session, agent, mode_id):
            del session, agent, mode_id
            return ModeState(
                current_mode_id="review",
                modes=[SessionMode(id="review", name="Review")],
            )

    model_agent = Agent(TestModel(custom_output_text="model"))
    model_adapter = _adapter(
        agent=model_agent,
        config=AdapterConfig(
            models_provider=EmptyModelsProvider(),
            session_store=MemorySessionStore(),
        ),
    )
    model_response = asyncio.run(model_adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    assert (
        asyncio.run(model_adapter.set_session_model("openai:gpt-5.4", model_response.session_id))
        is None
    )

    mode_agent = Agent(TestModel(custom_output_text="mode"))
    mode_adapter = _adapter(
        agent=mode_agent,
        config=AdapterConfig(
            modes_provider=ReviewModesProvider(),
            session_store=MemorySessionStore(),
        ),
    )
    mode_response = asyncio.run(mode_adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    assert (
        asyncio.run(mode_adapter.set_config_option("mode", mode_response.session_id, "review"))
        is not None
    )



from __future__ import annotations as _annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Protocol

from acp.interfaces import Client as AcpClient
from acp.schema import ReadTextFileResponse, WriteTextFileResponse

from ..session.state import AcpSessionContext
from .policy import HostAccessPolicy

__all__ = (
    "ClientFilesystemBackend",
    "FilesystemBackend",
)


class FilesystemBackend(Protocol):
    async def read_text_file(
        self,
        path: str,
        *,
        limit: int | None = None,
        line: int | None = None,
    ) -> ReadTextFileResponse: ...

    async def write_text_file(
        self,
        path: str,
        content: str,
    ) -> WriteTextFileResponse | None: ...


@dataclass(slots=True, frozen=True, kw_only=True)
class ClientFilesystemBackend:
    client: AcpClient
    session: AcpSessionContext
    access_policy: HostAccessPolicy | None = None
    workspace_root: Path | None = None

    async def read_text_file(
        self,
        path: str,
        *,
        limit: int | None = None,
        line: int | None = None,
    ) -> ReadTextFileResponse:
        self._enforce_path(path)
        return await self.client.read_text_file(
            path=path,
            session_id=self.session.session_id,
            limit=limit,
            line=line,
        )

    async def write_text_file(
        self,
        path: str,
        content: str,
    ) -> WriteTextFileResponse | None:
        self._enforce_path(path)
        return await self.client.write_text_file(
            content=content,
            path=path,
            session_id=self.session.session_id,
        )

    def _enforce_path(self, path: str) -> None:
        if self.access_policy is None:
            return
        self.access_policy.enforce_path(
            path,
            session_cwd=self.session.cwd,
            workspace_root=self.workspace_root,
        )



from __future__ import annotations as _annotations

from .state import AcpSessionContext
from .store import FileSessionStore, MemorySessionStore, SessionStore

__all__ = (
    "AcpSessionContext",
    "FileSessionStore",
    "MemorySessionStore",
    "SessionStore",
)



from __future__ import annotations as _annotations

import asyncio
from typing import Any, cast

import pytest
from pydantic_acp import RecordingACPClient, UpdateRecord, agent_message_texts

from .support import (
    AgentMessageChunk,
    AllowedOutcome,
    DeniedOutcome,
    PermissionOption,
    TerminalOutputResponse,
    ToolCallProgress,
    ToolCallStart,
    WaitForTerminalExitResponse,
    text_block,
)


def test_recording_acp_client_records_permission_file_and_terminal_calls() -> None:
    client = RecordingACPClient(
        write_response=None,
        release_response=None,
        kill_response=None,
        wait_response=WaitForTerminalExitResponse(exit_code=7),
        terminal_output_response=TerminalOutputResponse(output="captured", truncated=True),
    )
    permission_options = [
        PermissionOption(option_id="allow_once", name="Allow once", kind="allow_once"),
    ]
    tool_call = ToolCallProgress(
        session_update="tool_call_update",
        tool_call_id="tool-1",
        status="in_progress",
    )

    async def exercise_client() -> None:
        client.queue_permission_selected("allow_once")
        selected = await client.request_permission(
            permission_options,
            session_id="session-1",
            tool_call=tool_call,
        )
        client.queue_permission_cancelled()
        cancelled = await client.request_permission(
            permission_options,
            session_id="session-1",
            tool_call=tool_call,
        )

        assert isinstance(selected.outcome, AllowedOutcome)
        assert selected.outcome.option_id == "allow_once"
        assert isinstance(cancelled.outcome, DeniedOutcome)

        with pytest.raises(AssertionError, match="unexpected permission request"):
            await client.request_permission(
                permission_options,
                session_id="session-1",
                tool_call=tool_call,
            )

        await client.session_update(
            "session-1",
            AgentMessageChunk(
                session_update="agent_message_chunk",
                message_id="msg-1",
                content=text_block("hello"),
            ),
        )

        assert (await client.write_text_file("hello", "notes.txt", "session-1")) is None
        assert (
            await client.read_text_file("notes.txt", "session-1", limit=10, line=2)
        ).content == "file:notes.txt:2:10"
        assert (
            await client.create_terminal(
                "python",
                "session-1",
                args=["-V"],
                cwd="/tmp",
                output_byte_limit=5,
            )
        ).terminal_id == "terminal-1"
        assert (await client.terminal_output("session-1", "terminal-1")).output == "captured"
        assert (await client.release_terminal("session-1", "terminal-1")) is None
        assert (await client.wait_for_terminal_exit("session-1", "terminal-1")).exit_code == 7
        assert (await client.kill_terminal("session-1", "terminal-1")) is None

        with pytest.raises(AssertionError, match="unexpected extension method"):
            await client.ext_method("ext/test", {})
        with pytest.raises(AssertionError, match="unexpected extension notification"):
            await client.ext_notification("ext/test", {})

    asyncio.run(exercise_client())
    client.on_connect(cast("Any", object()))

    assert client.permission_option_ids == [
        ("session-1", ["allow_once"], tool_call),
        ("session-1", ["allow_once"], tool_call),
        ("session-1", ["allow_once"], tool_call),
    ]
    assert len(client.updates) == 1
    assert client.write_calls == [("session-1", "notes.txt", "hello")]
    assert client.read_calls == [("session-1", "notes.txt", 10, 2)]
    assert client.create_calls == [
        ("session-1", "python", ["-V"], "/tmp", None, 5),
    ]
    assert client.output_calls == [("session-1", "terminal-1")]
    assert client.release_calls == [("session-1", "terminal-1")]
    assert client.wait_calls == [("session-1", "terminal-1")]
    assert client.kill_calls == [("session-1", "terminal-1")]


def test_agent_message_texts_groups_streamed_and_anonymous_messages() -> None:
    client = RecordingACPClient(
        updates=[
            UpdateRecord(
                session_id="session-1",
                update=ToolCallStart(
                    session_update="tool_call",
                    tool_call_id="tool-1",
                    title="tool",
                    kind="execute",
                    status="in_progress",
                ),
            ),
            UpdateRecord(
                session_id="session-1",
                update=AgentMessageChunk(
                    session_update="agent_message_chunk",
                    message_id="msg-1",
                    content=text_block("hel"),
                ),
            ),
            UpdateRecord(
                session_id="session-1",
                update=AgentMessageChunk(
                    session_update="agent_message_chunk",
                    message_id="msg-1",
                    content=text_block("lo"),
                ),
            ),
            UpdateRecord(
                session_id="session-1",
                update=AgentMessageChunk(
                    session_update="agent_message_chunk",
                    message_id=None,
                    content=text_block("anon"),
                ),
            ),
            UpdateRecord(
                session_id="session-1",
                update=AgentMessageChunk(
                    session_update="agent_message_chunk",
                    message_id=None,
                    content=text_block("second"),
                ),
            ),
        ],
    )

    assert agent_message_texts(client) == ["hello", "anon", "second"]
    assert agent_message_texts(RecordingACPClient()) == []



from __future__ import annotations as _annotations

from dataclasses import dataclass
from typing import TypeAlias

from acp.schema import ModelInfo
from pydantic_ai.models import Model as PydanticModel

ModelOverride: TypeAlias = str | PydanticModel

__all__ = ("AdapterModel", "ModelOverride")


@dataclass(slots=True, frozen=True, kw_only=True)
class AdapterModel:
    model_id: str
    name: str
    override: ModelOverride
    description: str | None = None

    def to_model_info(self) -> ModelInfo:
        return ModelInfo(
            model_id=self.model_id,
            name=self.name,
            description=self.description,
        )



from __future__ import annotations as _annotations

from collections.abc import Sequence
from pathlib import Path

from ._policy_paths import normalize_host_path

__all__ = ("extract_command_path_candidates", "resolve_command_cwd")


def resolve_command_cwd(
    cwd: str | Path | None,
    *,
    session_cwd: Path,
) -> Path:
    if cwd is None:
        return session_cwd.resolve(strict=False)
    return normalize_host_path(cwd, base_dir=session_cwd)[0]


def extract_command_path_candidates(
    command: str,
    *,
    args: Sequence[str] | None,
    command_cwd: Path,
) -> tuple[Path, ...]:
    candidates: list[Path] = []
    _append_path_candidate(candidates, command, base_dir=command_cwd)
    for arg in args or ():
        token = _extract_path_token(arg)
        if token is None:
            continue
        _append_path_candidate(candidates, token, base_dir=command_cwd)
    return tuple(_dedupe_paths(candidates))


def _append_path_candidate(
    candidates: list[Path],
    token: str,
    *,
    base_dir: Path,
) -> None:
    if not _looks_like_path_token(token):
        return
    candidates.append(normalize_host_path(token, base_dir=base_dir)[0])


def _extract_path_token(token: str) -> str | None:
    if _looks_like_path_token(token):
        return token
    if token.startswith("--") and "=" in token:
        _, value = token.split("=", 1)
        if _looks_like_path_token(value):
            return value
    if "=" in token and not token.startswith("-"):
        _, value = token.split("=", 1)
        if _looks_like_path_token(value):
            return value
    return None


def _looks_like_path_token(token: str) -> bool:
    if not token:
        return False
    if token in {".", ".."}:
        return True
    if token.startswith("-"):
        return False
    return (
        token.startswith(("/", "./", "../", "~/", ".\\", "..\\")) or "/" in token or "\\" in token
    )


def _dedupe_paths(paths: Sequence[Path]) -> list[Path]:
    seen: set[Path] = set()
    deduped: list[Path] = []
    for path in paths:
        if path in seen:
            continue
        seen.add(path)
        deduped.append(path)
    return deduped



from __future__ import annotations as _annotations

import pytest

from acpkit import CompatibilityManifest, SurfaceSupport


def test_compatibility_manifest_validates_and_renders_markdown() -> None:
    manifest = CompatibilityManifest(
        integration_name="workspace-agent",
        adapter="pydantic-acp",
        surfaces={
            "session.load": SurfaceSupport(
                status="implemented",
                owner="adapter",
                mapping="FileSessionStore + load_session",
            ),
            "mode.switch": SurfaceSupport(
                status="partial",
                owner="bridge",
                mapping="PrepareToolsBridge dynamic modes",
                rationale="Only explicitly exposed runtime modes are surfaced.",
            ),
            "authenticate": SurfaceSupport(
                status="planned",
                rationale="No auth handshake has been added yet.",
            ),
        },
    )

    manifest.validate()
    markdown = manifest.to_markdown()

    assert "# Compatibility Manifest: workspace-agent" in markdown
    assert (
        "| session.load | implemented | adapter | FileSessionStore + load_session |  |" in markdown
    )
    assert "| authenticate | planned |  |  | No auth handshake has been added yet. |" in markdown


def test_compatibility_manifest_requires_owner_for_implemented_surface() -> None:
    manifest = CompatibilityManifest(
        integration_name="workspace-agent",
        adapter="pydantic-acp",
        surfaces={
            "session.load": SurfaceSupport(
                status="implemented",
                mapping="FileSessionStore + load_session",
            ),
        },
    )

    with pytest.raises(ValueError, match="must declare an owner"):
        manifest.validate()


def test_compatibility_manifest_requires_mapping_for_implemented_surface() -> None:
    manifest = CompatibilityManifest(
        integration_name="workspace-agent",
        adapter="pydantic-acp",
        surfaces={
            "session.load": SurfaceSupport(
                status="implemented",
                owner="adapter",
            ),
        },
    )

    with pytest.raises(ValueError, match="must declare a mapping"):
        manifest.validate()


def test_compatibility_manifest_requires_rationale_for_intentionally_unused_surface() -> None:
    manifest = CompatibilityManifest(
        integration_name="workspace-agent",
        adapter="pydantic-acp",
        surfaces={
            "hooks.visible": SurfaceSupport(
                status="intentionally_not_used",
            ),
        },
    )

    with pytest.raises(ValueError, match="must declare a rationale"):
        manifest.validate()


def test_compatibility_manifest_requires_rationale_for_mixed_owner() -> None:
    manifest = CompatibilityManifest(
        integration_name="workspace-agent",
        adapter="pydantic-acp",
        surfaces={
            "model.switch": SurfaceSupport(
                status="partial",
                owner="mixed",
                mapping="Built-in selection plus provider override",
                rationale="",
            ),
        },
    )

    with pytest.raises(ValueError, match="must declare a rationale"):
        manifest.validate()


def test_compatibility_manifest_mixed_owner_requires_explanation_even_when_implemented() -> None:
    manifest = CompatibilityManifest(
        integration_name="workspace-agent",
        adapter="pydantic-acp",
        surfaces={
            "session.sync": SurfaceSupport(
                status="implemented",
                owner="mixed",
                mapping="adapter plus host persistence",
            ),
        },
    )

    with pytest.raises(ValueError, match="explain how mixed ownership works"):
        manifest.validate()


def test_compatibility_manifest_rejects_blank_headers_and_whitespace_surface_names() -> None:
    with pytest.raises(ValueError, match="integration_name"):
        CompatibilityManifest(
            integration_name=" ",
            adapter="pydantic-acp",
            surfaces={"session.load": SurfaceSupport(status="planned", rationale="later")},
        ).validate()

    with pytest.raises(ValueError, match="adapter"):
        CompatibilityManifest(
            integration_name="workspace-agent",
            adapter=" ",
            surfaces={"session.load": SurfaceSupport(status="planned", rationale="later")},
        ).validate()

    with pytest.raises(ValueError, match="at least one declared surface"):
        CompatibilityManifest(
            integration_name="workspace-agent",
            adapter="pydantic-acp",
            surfaces={},
        ).validate()

    with pytest.raises(ValueError, match="must not be empty"):
        CompatibilityManifest(
            integration_name="workspace-agent",
            adapter="pydantic-acp",
            surfaces={"": SurfaceSupport(status="planned", rationale="later")},
        ).validate()

    with pytest.raises(ValueError, match="surrounding whitespace"):
        CompatibilityManifest(
            integration_name="workspace-agent",
            adapter="pydantic-acp",
            surfaces={" session.load ": SurfaceSupport(status="planned", rationale="later")},
        ).validate()


def test_compatibility_manifest_partial_surface_requires_owner_and_blank_cells_render_empty() -> (
    None
):
    manifest = CompatibilityManifest(
        integration_name="workspace-agent",
        adapter="pydantic-acp",
        surfaces={
            "mode.switch": SurfaceSupport(
                status="partial",
                rationale="Host chooses which modes stay visible.",
            ),
            "notes": SurfaceSupport(
                status="planned",
                rationale="future work",
                mapping="",
            ),
        },
    )

    with pytest.raises(ValueError, match="must declare an owner"):
        manifest.validate()

    fixed_manifest = CompatibilityManifest(
        integration_name="workspace-agent",
        adapter="pydantic-acp",
        surfaces={
            "mode.switch": SurfaceSupport(
                status="partial",
                owner="bridge",
                rationale="Host chooses which modes stay visible.",
            ),
            "notes": SurfaceSupport(
                status="planned",
                rationale="future work",
                mapping="",
            ),
        },
    )

    markdown = fixed_manifest.to_markdown()
    assert "| notes | planned |  |  | future work |" in markdown



from __future__ import annotations as _annotations

from dataclasses import dataclass

from ._version import __version__
from .config import ServerOptions

__all__ = ("ServerMetadata", "TransportMetadata", "build_server_metadata")


@dataclass(frozen=True, kw_only=True)
class TransportMetadata:
    transport_kind: str = "websocket"
    transport_version: int = 1
    package_version: str = __version__


@dataclass(frozen=True, kw_only=True)
class ServerMetadata:
    transport_kind: str
    transport_version: int
    package_version: str
    auth_required: bool
    supported_auth_modes: tuple[str, ...]
    max_size: int
    max_queue: int
    compression: str | None
    health_path: str
    metadata_path: str
    websocket_path: str
    supported_agent_families: tuple[str, ...] = ()
    remote_cwd: str | None = None

    def to_json_dict(self) -> dict[str, object]:
        return {
            "transport_kind": self.transport_kind,
            "transport_version": self.transport_version,
            "package_version": self.package_version,
            "auth_required": self.auth_required,
            "supported_auth_modes": list(self.supported_auth_modes),
            "max_size": self.max_size,
            "max_queue": self.max_queue,
            "compression": self.compression,
            "health_path": self.health_path,
            "metadata_path": self.metadata_path,
            "websocket_path": self.websocket_path,
            "supported_agent_families": list(self.supported_agent_families),
            "remote_cwd": self.remote_cwd,
        }


def build_server_metadata(options: ServerOptions) -> ServerMetadata:
    transport_metadata = TransportMetadata()
    auth_required = options.bearer_token is not None and bool(options.bearer_token.strip())
    return ServerMetadata(
        transport_kind=transport_metadata.transport_kind,
        transport_version=transport_metadata.transport_version,
        package_version=transport_metadata.package_version,
        auth_required=auth_required,
        supported_auth_modes=("bearer",) if auth_required else (),
        max_size=options.transport.max_size,
        max_queue=options.transport.max_queue,
        compression=options.transport.compression,
        health_path=options.paths.health_path,
        metadata_path=options.paths.metadata_path,
        websocket_path=options.paths.websocket_path,
        supported_agent_families=options.supported_agent_families,
        remote_cwd=options.remote_cwd,
    )



from __future__ import annotations as _annotations

from pathlib import Path
from typing import Any, cast

import pytest
from acp.schema import AgentMessageChunk, PermissionOption, RequestPermissionResponse
from pydantic_ai import Agent
from pydantic_ai.models.test import TestModel

from .support import (
    UTC,
    AcpSessionContext,
    AsyncDemoConfigOptionsProvider,
    AsyncDemoModesProvider,
    AsyncDemoPlanProvider,
    DemoConfigOptionsProvider,
    DemoModelsProvider,
    DemoModesProvider,
    FilesystemRecordingClient,
    FreeformModelsProvider,
    HostRecordingClient,
    RecordingClient,
    ReservedModelConfigProvider,
    TerminalRecordingClient,
    agent_message_texts,
    datetime,
    text_block,
)


def _session() -> AcpSessionContext:
    created_at = datetime.now(UTC)
    return AcpSessionContext(
        session_id="session-1",
        cwd=Path("/tmp/demo"),
        created_at=created_at,
        updated_at=created_at,
    )


@pytest.mark.asyncio
async def test_pydantic_support_recording_clients_cover_helpers_and_error_paths() -> None:
    client = RecordingClient()
    option = PermissionOption(option_id="allow_once", name="Allow once", kind="allow_once")
    tool_call = cast("Any", object())

    with pytest.raises(AssertionError, match="unexpected permission request"):
        await client.request_permission([option], "session-1", tool_call)

    client.queue_permission_selected("allow_once")
    selected = await client.request_permission([option], "session-1", tool_call)
    assert isinstance(selected, RequestPermissionResponse)

    client.queue_permission_cancelled()
    cancelled = await client.request_permission([option], "session-1", tool_call)
    assert isinstance(cancelled, RequestPermissionResponse)

    update = AgentMessageChunk(
        session_update="agent_message_chunk",
        message_id=None,
        content=text_block("hello"),
    )
    await client.session_update("session-1", update)
    client.updates.append(("session-1", object()))
    assert agent_message_texts(client) == ["hello"]

    with pytest.raises(AssertionError, match="filesystem flow"):
        await client.write_text_file("body", "/tmp/demo", "session-1")
    with pytest.raises(AssertionError, match="filesystem flow"):
        await client.read_text_file("/tmp/demo", "session-1")
    with pytest.raises(AssertionError, match="terminal flow"):
        await client.create_terminal("echo hi", "session-1")
    with pytest.raises(AssertionError, match="terminal flow"):
        await client.terminal_output("session-1", "terminal-1")
    with pytest.raises(AssertionError, match="terminal flow"):
        await client.release_terminal("session-1", "terminal-1")
    with pytest.raises(AssertionError, match="terminal flow"):
        await client.wait_for_terminal_exit("session-1", "terminal-1")
    with pytest.raises(AssertionError, match="terminal flow"):
        await client.kill_terminal("session-1", "terminal-1")
    with pytest.raises(AssertionError, match="unexpected extension method"):
        await client.ext_method("demo.echo", {"value": 1})
    with pytest.raises(AssertionError, match="unexpected extension notification"):
        await client.ext_notification("demo.note", {"value": 2})
    assert client.on_connect(cast("Any", object())) is None

    fs_client = FilesystemRecordingClient()
    assert (
        await fs_client.write_text_file("body", "/tmp/demo", "session-1")
        == fs_client.write_response
    )
    read_response = await fs_client.read_text_file("/tmp/demo", "session-1", limit=5, line=2)
    assert read_response.content == "file:/tmp/demo:2:5"

    terminal_client = TerminalRecordingClient()
    created = await terminal_client.create_terminal("echo hi", "session-1")
    assert created.terminal_id == "terminal-1"
    assert (
        await terminal_client.terminal_output("session-1", "terminal-1")
    ).output == "terminal-output"
    assert (
        await terminal_client.release_terminal("session-1", "terminal-1")
        == terminal_client.release_response
    )
    assert (
        await terminal_client.wait_for_terminal_exit("session-1", "terminal-1")
        == terminal_client.wait_response
    )
    assert (
        await terminal_client.kill_terminal("session-1", "terminal-1")
        == terminal_client.kill_response
    )

    host_client = HostRecordingClient()
    await host_client.write_text_file("body", "/tmp/demo", "session-1")
    await host_client.create_terminal("echo hi", "session-1")
    assert host_client.write_calls == [("session-1", "/tmp/demo", "body")]
    assert host_client.create_calls[0][1] == "echo hi"


@pytest.mark.asyncio
async def test_pydantic_support_providers_cover_valid_and_invalid_paths() -> None:
    session = _session()
    agent = Agent(TestModel())

    models = DemoModelsProvider()
    model_state = await models.get_model_state(session, agent)
    assert model_state.current_model_id == "provider-model-a"
    with pytest.raises(AssertionError, match="unexpected model id"):
        await models.set_model(session, agent, "missing-model")
    updated_model_state = await models.set_model(session, agent, "provider-model-b")
    assert updated_model_state.current_model_id == "provider-model-b"

    modes = DemoModesProvider()
    mode_state = modes.get_mode_state(session, agent)
    assert mode_state.current_mode_id == "chat"
    with pytest.raises(AssertionError, match="unexpected mode id"):
        modes.set_mode(session, agent, "missing-mode")
    updated_mode_state = modes.set_mode(session, agent, "review")
    assert updated_mode_state.current_mode_id == "review"

    config_options = DemoConfigOptionsProvider()
    assert config_options.set_config_option(session, agent, "wrong", True) is None
    assert config_options.set_config_option(session, agent, "stream_enabled", "yes") is None
    updated_options = config_options.set_config_option(session, agent, "stream_enabled", True)
    assert updated_options is not None

    freeform = FreeformModelsProvider()
    freeform_state = await freeform.set_model(session, agent, "custom-model-b")
    assert freeform_state.current_model_id == "custom-model-b"

    reserved = ReservedModelConfigProvider()
    assert reserved.set_config_option(session, agent, "wrong", "value") is None
    updated_reserved = reserved.set_config_option(session, agent, "model", "provider-model-b")
    assert updated_reserved is not None

    async_modes = AsyncDemoModesProvider()
    assert (await async_modes.get_mode_state(session, agent)).current_mode_id == "review"
    assert (await async_modes.set_mode(session, agent, "chat")).current_mode_id == "chat"

    async_config = AsyncDemoConfigOptionsProvider()
    assert await async_config.get_config_options(session, agent)
    assert await async_config.set_config_option(session, agent, "stream_enabled", False) is not None

    async_plan = AsyncDemoPlanProvider()
    plan = await async_plan.get_plan(session, agent)
    assert [entry.content for entry in plan] == ["mode:chat", "stream:false"]



from __future__ import annotations as _annotations

from .adapter import PydanticAcpAgent
from .server import create_acp_agent, run_acp

__all__ = (
    "PydanticAcpAgent",
    "create_acp_agent",
    "run_acp",
)



from __future__ import annotations as _annotations

from typing import Final

__all__ = (
    "DEFAULT_CLOSE_TIMEOUT",
    "DEFAULT_MAX_MESSAGE_SIZE",
    "DEFAULT_MAX_QUEUE",
    "DEFAULT_OPEN_TIMEOUT",
    "DEFAULT_PING_INTERVAL",
    "DEFAULT_PING_TIMEOUT",
)

DEFAULT_MAX_MESSAGE_SIZE: Final[int] = 1_048_576
DEFAULT_MAX_QUEUE: Final[int] = 16
DEFAULT_OPEN_TIMEOUT: Final[float] = 10.0
DEFAULT_PING_INTERVAL: Final[float] = 20.0
DEFAULT_PING_TIMEOUT: Final[float] = 20.0
DEFAULT_CLOSE_TIMEOUT: Final[float] = 10.0



from __future__ import annotations as _annotations

import asyncio
import os
import shlex

from acpremote import serve_command


def _command() -> tuple[str, ...]:
    raw = os.getenv("ACPREMOTE_COMMAND", "").strip()
    if raw:
        return tuple(shlex.split(raw))
    return (
        "acpkit",
        "run",
        "examples.langchain.workspace_graph:graph",
    )


def _host() -> str:
    return os.getenv("ACPREMOTE_HOST", "127.0.0.1")


def _port() -> int:
    return int(os.getenv("ACPREMOTE_PORT", "8080"))


async def main() -> None:
    server = await serve_command(
        _command(),
        host=_host(),
        port=_port(),
    )
    await server.serve_forever()


if __name__ == "__main__":
    asyncio.run(main())



from __future__ import annotations as _annotations

import sys
from collections.abc import Sequence

import click

from .runtime import (
    AcpKitError,
    launch_target,
    run_remote_addr,
    run_target,
    serve_target,
)
from .runtime import launch_command as launch_raw_command

__all__ = ("cli", "main")


class AcpKitClickError(click.ClickException):
    exit_code = 2


@click.group(context_settings={"help_option_names": ["-h", "--help"]})
def cli() -> None:
    """ACP Kit root CLI."""


@cli.command("run")
@click.argument("target", required=False)
@click.option(
    "--addr",
    help="Remote ACP WebSocket address to expose locally over stdio.",
)
@click.option(
    "--token-env",
    help="Environment variable containing the bearer token for `--addr`.",
)
@click.option(
    "-p",
    "--path",
    "import_roots",
    multiple=True,
    help="Extra import root for loading the target module. Can be repeated.",
)
def run_command(
    target: str | None,
    addr: str | None,
    token_env: str | None,
    import_roots: tuple[str, ...],
) -> None:
    if (target is None) == (addr is None):
        raise AcpKitClickError("Provide exactly one of `TARGET` or `--addr`.")
    if addr is not None and import_roots:
        raise AcpKitClickError("`--path` can only be used with `TARGET`.")
    try:
        if addr is not None:
            run_remote_addr(addr, token_env=token_env)
        else:
            assert target is not None
            run_target(target, import_roots=import_roots)
    except AcpKitError as exc:
        raise AcpKitClickError(str(exc)) from exc


@cli.command("serve")
@click.argument("target")
@click.option(
    "--host",
    default="127.0.0.1",
    show_default=True,
    help="Host interface for the remote ACP WebSocket server.",
)
@click.option(
    "--port",
    default=8080,
    show_default=True,
    type=int,
    help="TCP port for the remote ACP WebSocket server.",
)
@click.option(
    "--mount-path",
    default="/acp",
    show_default=True,
    help="Mount path for metadata and WebSocket routes.",
)
@click.option(
    "--token-env",
    help="Environment variable containing the bearer token to require.",
)
@click.option(
    "-p",
    "--path",
    "import_roots",
    multiple=True,
    help="Extra import root for loading the target module. Can be repeated.",
)
def serve_command(
    target: str,
    host: str,
    port: int,
    mount_path: str,
    token_env: str | None,
    import_roots: tuple[str, ...],
) -> None:
    try:
        serve_target(
            target,
            import_roots=import_roots,
            host=host,
            port=port,
            mount_path=mount_path,
            token_env=token_env,
        )
    except AcpKitError as exc:
        raise AcpKitClickError(str(exc)) from exc


@cli.command("launch")
@click.argument("target", required=False)
@click.option(
    "-c",
    "--command",
    "raw_command",
    help="Launch a raw server command directly with `toad acp`, for example `python3.11 finance_agent.py`.",
)
@click.option(
    "-p",
    "--path",
    "import_roots",
    multiple=True,
    help="Extra import root for loading the target module. Can be repeated.",
)
def launch_command(
    target: str | None,
    raw_command: str | None,
    import_roots: tuple[str, ...],
) -> None:
    if (target is None) == (raw_command is None):
        raise AcpKitClickError("Provide exactly one of `TARGET` or `--command`.")
    if raw_command is not None and import_roots:
        raise AcpKitClickError("`--path` can only be used when launching a target.")
    try:
        if raw_command is not None:
            exit_code = launch_raw_command(raw_command)
        else:
            assert target is not None
            exit_code = launch_target(target, import_roots=import_roots)
    except AcpKitError as exc:
        raise AcpKitClickError(str(exc)) from exc
    if exit_code != 0:
        raise click.exceptions.Exit(exit_code)


def main(argv: Sequence[str] | None = None) -> int:
    try:
        cli.main(
            args=list(argv) if argv is not None else None,
            prog_name="acpkit",
            standalone_mode=False,
        )
    except click.ClickException as exc:
        click.echo(f"Error: {exc.format_message()}", file=sys.stderr, err=True)
        return exc.exit_code
    except click.exceptions.Exit as exc:
        return exc.exit_code
    return 0  # pragma: no cover



from __future__ import annotations as _annotations

from contextvars import ContextVar
from dataclasses import dataclass
from typing import Any, cast

from pydantic_acp.runtime._compat import (
    build_capability_override_value,
    entry_func,
    entry_timeout,
    entry_tool_filters,
    hook_registry,
    replace_hook_entry,
    root_capability,
    root_capability_override,
)
from pydantic_ai import Agent
from pydantic_ai.capabilities import CombinedCapability, Hooks
from pydantic_ai.models.test import TestModel


@dataclass
class _DataclassEntry:
    func: Any
    timeout: float | None
    tools: frozenset[str] | None = None


@dataclass
class _IncompatibleDataclassEntry:
    func: Any


class _MutableEntry:
    def __init__(self, func: Any, timeout: float | None) -> None:
        self.func = func
        self.timeout = timeout
        self.tools = frozenset({"b", "a"})


class _UncopyableEntry:
    def __copy__(self) -> _UncopyableEntry:
        raise RuntimeError("no copy")


class _ReadOnlyEntry:
    __slots__ = ("_func", "_timeout")

    def __init__(self, func: Any, timeout: float | None) -> None:
        self._func = func
        self._timeout = timeout

    @property
    def func(self) -> Any:
        return self._func  # pragma: no cover

    @property
    def timeout(self) -> float | None:
        return self._timeout  # pragma: no cover


def test_root_capability_and_override_helpers_handle_missing_and_present_private_attrs() -> None:
    agent = Agent(TestModel(custom_output_text="ok"))
    capability = CombinedCapability(capabilities=[])
    override = ContextVar[object]("override")

    cast("Any", agent)._root_capability = capability
    cast("Any", agent)._override_root_capability = override

    override_value = build_capability_override_value(capability)

    assert root_capability(agent) is capability
    assert root_capability_override(agent) is override
    assert type(override_value).__name__ == "_CapabilityOverrideValue"
    assert cast("Any", override_value).value is capability

    cast("Any", agent)._root_capability = "wrong"
    cast("Any", agent)._override_root_capability = "wrong"
    assert root_capability(agent) is None
    assert root_capability_override(agent) is None


def test_hook_registry_and_entry_helpers_normalize_supported_shapes() -> None:
    hooks = Hooks[Any]()
    valid_entry = _DataclassEntry(
        func="callable",
        timeout=1,
        tools=cast("Any", frozenset({"b", "a", 1})),
    )
    cast("Any", hooks)._registry = {
        "before_run": [valid_entry],
        1: [valid_entry],
        "bad": "wrong",
    }

    registry = hook_registry(hooks)

    assert registry == {"before_run": [valid_entry]}
    assert entry_func(valid_entry) == "callable"
    assert entry_timeout(valid_entry) == 1.0
    assert entry_tool_filters(valid_entry) == ("a", "b")
    assert entry_timeout(object()) is None
    assert entry_tool_filters(object()) == ()

    cast("Any", hooks)._registry = "wrong"
    assert hook_registry(hooks) is None


def test_replace_hook_entry_supports_dataclass_copy_and_failure_paths() -> None:
    dataclass_entry = _DataclassEntry(func="old", timeout=1.0)
    replaced_dataclass = replace_hook_entry(dataclass_entry, func="new", timeout=2.0)
    assert isinstance(replaced_dataclass, _DataclassEntry)
    assert replaced_dataclass.func == "new"
    assert replaced_dataclass.timeout == 2.0

    incompatible_dataclass = _IncompatibleDataclassEntry(func="old")
    assert replace_hook_entry(incompatible_dataclass, func="new", timeout=2.0) is None

    mutable_entry = _MutableEntry(func="old", timeout=1.0)
    replaced_mutable = replace_hook_entry(mutable_entry, func="new", timeout=2.0)
    assert isinstance(replaced_mutable, _MutableEntry)
    assert replaced_mutable is not mutable_entry
    assert replaced_mutable.func == "new"
    assert replaced_mutable.timeout == 2.0

    assert replace_hook_entry(_UncopyableEntry(), func="new", timeout=2.0) is None
    assert (
        replace_hook_entry(_ReadOnlyEntry(func="old", timeout=1.0), func="new", timeout=2.0) is None
    )



from __future__ import annotations as _annotations

from collections.abc import Awaitable, Callable
from dataclasses import dataclass, field
from typing import TypeAlias, TypeVar

from pydantic_ai.messages import ModelMessage
from pydantic_ai.tools import RunContext

from ..agent_types import RuntimeAgent
from ..awaitables import resolve_value
from ..session.state import AcpSessionContext, JsonValue
from .base import BufferedCapabilityBridge

AgentDepsT = TypeVar("AgentDepsT", contravariant=True)
ModelMessages: TypeAlias = list[ModelMessage]
HistoryProcessorNoContextSync: TypeAlias = Callable[[ModelMessages], ModelMessages]
HistoryProcessorNoContextAsync: TypeAlias = Callable[[ModelMessages], Awaitable[ModelMessages]]
HistoryProcessorWithContextSync: TypeAlias = Callable[
    [RunContext[AgentDepsT], ModelMessages],
    ModelMessages,
]
HistoryProcessorWithContextAsync: TypeAlias = Callable[
    [RunContext[AgentDepsT], ModelMessages],
    Awaitable[ModelMessages],
]
HistoryProcessorPlain: TypeAlias = HistoryProcessorNoContextSync | HistoryProcessorNoContextAsync
HistoryProcessorContextual: TypeAlias = (
    HistoryProcessorWithContextSync[AgentDepsT] | HistoryProcessorWithContextAsync[AgentDepsT]
)
HistoryProcessorCallable: TypeAlias = HistoryProcessorPlain | HistoryProcessorContextual[AgentDepsT]

__all__ = (
    "HistoryProcessorBridge",
    "HistoryProcessorCallable",
    "HistoryProcessorContextual",
    "HistoryProcessorPlain",
    "HistoryProcessorWithContextAsync",
    "HistoryProcessorWithContextSync",
)


@dataclass(slots=True)
class HistoryProcessorBridge(BufferedCapabilityBridge):
    metadata_key: str | None = "history_processors"
    processor_names: list[str] = field(default_factory=list)

    def wrap_plain_processor(
        self,
        session: AcpSessionContext,
        processor: HistoryProcessorPlain,
        *,
        name: str = "history_processor",
    ) -> HistoryProcessorNoContextAsync:
        self._register_processor_name(name)

        async def wrapped(messages: ModelMessages) -> ModelMessages:
            return await self._run_plain_processor(
                session,
                processor,
                name=name,
                messages=messages,
            )

        return wrapped

    def wrap_contextual_processor(
        self,
        session: AcpSessionContext,
        processor: HistoryProcessorContextual[AgentDepsT],
        *,
        name: str = "history_processor",
    ) -> HistoryProcessorWithContextAsync[AgentDepsT]:
        self._register_processor_name(name)

        async def wrapped(
            ctx: RunContext[AgentDepsT],
            messages: ModelMessages,
        ) -> ModelMessages:
            return await self._run_contextual_processor(
                session,
                processor,
                name=name,
                messages=messages,
                ctx=ctx,
            )

        return wrapped

    def get_session_metadata(
        self,
        session: AcpSessionContext,
        agent: RuntimeAgent,
    ) -> dict[str, JsonValue]:
        del session, agent
        processors: list[JsonValue] = list(self.processor_names)
        return {"processors": processors}

    def _register_processor_name(self, name: str) -> None:
        if name not in self.processor_names:
            self.processor_names.append(name)

    async def _run_contextual_processor(
        self,
        session: AcpSessionContext,
        processor: HistoryProcessorContextual[AgentDepsT],
        *,
        name: str,
        ctx: RunContext[AgentDepsT],
        messages: ModelMessages,
    ) -> ModelMessages:
        try:
            result = processor(ctx, messages)
            processed = await resolve_value(result)
        except Exception as error:
            self._record_failed_event(
                session,
                title=f"history_processor.{name}",
                raw_output=str(error),
            )
            raise

        self._record_completed_event(
            session,
            title=f"history_processor.{name}",
            raw_output=f"messages={len(messages)}->{len(processed)}",
        )
        return processed

    async def _run_plain_processor(
        self,
        session: AcpSessionContext,
        processor: HistoryProcessorPlain,
        *,
        name: str,
        messages: ModelMessages,
    ) -> ModelMessages:
        try:
            result = processor(messages)
            processed = await resolve_value(result)
        except Exception as error:
            self._record_failed_event(
                session,
                title=f"history_processor.{name}",
                raw_output=str(error),
            )
            raise

        self._record_completed_event(
            session,
            title=f"history_processor.{name}",
            raw_output=f"messages={len(messages)}->{len(processed)}",
        )
        return processed



from __future__ import annotations as _annotations

import os
from collections.abc import Sequence
from pathlib import Path
from typing import Final

from pydantic_acp import (
    AcpSessionContext,
    AdapterConfig,
    FileSessionStore,
    FileSystemProjectionMap,
    HookProjectionMap,
    RuntimeAgent,
    create_acp_agent,
    run_acp,
)
from pydantic_acp.models import ModelOverride
from pydantic_acp.types import (
    AgentPromptBlock,
    AudioContentBlock,
    BlobResourceContents,
    EmbeddedResourceContentBlock,
    ImageContentBlock,
    ResourceContentBlock,
)
from pydantic_ai import Agent
from pydantic_ai.capabilities import Hooks, ValidatedToolArgs
from pydantic_ai.messages import ModelResponse, ToolCallPart
from pydantic_ai.models import ModelRequestContext
from pydantic_ai.models.test import TestModel
from pydantic_ai.tools import DeferredToolRequests, RunContext, ToolDefinition

__all__ = ("TravelPromptModelProvider", "acp_agent", "agent", "config", "main")

_DEMO_ROOT: Final[Path] = Path("agent_demos")
_TRAVEL_ROOT: Final[Path] = Path.cwd() / _DEMO_ROOT / "travel-agent"
_READ_TOOL: Final[str] = "read_trip_file"
_WRITE_TOOL: Final[str] = "write_trip_file"
_SESSION_STORE_ROOT: Final[Path] = (
    Path(os.getenv("ACP_EXAMPLE_SESSION_DIR", str(_DEMO_ROOT / "acp-sessions")))
    .expanduser()
    .resolve()
    / "pydantic-travel"
)
_MEDIA_MODEL_ENV_NAMES: Final[tuple[str, ...]] = (
    "ACP_TRAVEL_MEDIA_MODEL",
    "TRAVEL_MEDIA_MODEL",
)
_DEFAULT_FILES: Final[dict[str, str]] = {
    "itinerary.md": (
        "# Travel Brief\n\n"
        "- city: Lisbon\n"
        "- trip goal: design-heavy long weekend\n"
        "- constraints: walkable neighborhoods, low-friction transit\n"
    ),
    "ideas.txt": (
        "Try these prompts:\n"
        "- list trip files\n"
        "- read trip file itinerary.md\n"
        "- write trip file scratch.txt: book the riverside hotel\n"
    ),
}


def _default_model_name() -> str | TestModel:
    configured_model = os.getenv("ACP_TRAVEL_MODEL", "").strip()
    if configured_model:
        return configured_model
    return TestModel()


def _configured_media_model_name() -> str | None:
    for env_name in _MEDIA_MODEL_ENV_NAMES:
        configured_model = os.getenv(env_name, "").strip()
        if configured_model:
            return configured_model
    return None


def _ensure_travel_workspace() -> None:
    _TRAVEL_ROOT.mkdir(parents=True, exist_ok=True)
    for relative_path, content in _DEFAULT_FILES.items():
        file_path = _resolve_trip_path(relative_path)
        if not file_path.exists():
            file_path.write_text(content, encoding="utf-8")


def _resolve_trip_path(path: str) -> Path:
    candidate = (_TRAVEL_ROOT / path).resolve()
    try:
        candidate.relative_to(_TRAVEL_ROOT)
    except ValueError as exc:
        raise ValueError("Path must stay inside the travel demo workspace.") from exc
    return candidate


def _truncate_text(text: str, *, limit: int) -> str:
    if len(text) <= limit:
        return text
    return f"{text[:limit]}\n\n...[truncated]"


def _prompt_has_binary_media(prompt: Sequence[AgentPromptBlock]) -> bool:
    for block in prompt:
        if isinstance(block, ImageContentBlock | AudioContentBlock):
            return True
        if isinstance(block, ResourceContentBlock):
            mime_type = block.mime_type
            if mime_type is not None and not mime_type.startswith("text/"):
                return True
            continue
        if isinstance(block, EmbeddedResourceContentBlock) and isinstance(
            block.resource,
            BlobResourceContents,
        ):
            return True
    return False


def _prompt_has_image_media(prompt: Sequence[AgentPromptBlock]) -> bool:
    for block in prompt:
        if isinstance(block, ImageContentBlock):
            return True
        if isinstance(block, ResourceContentBlock):
            mime_type = block.mime_type
            if mime_type is not None and mime_type.startswith("image/"):
                return True
            continue
        if isinstance(block, EmbeddedResourceContentBlock):
            resource = block.resource
            if isinstance(resource, BlobResourceContents):
                mime_type = resource.mime_type
                if mime_type is not None and mime_type.startswith("image/"):
                    return True
    return False


class TravelPromptModelProvider:
    def get_prompt_model_override(
        self,
        session: AcpSessionContext,
        agent: RuntimeAgent,
        prompt: Sequence[AgentPromptBlock],
        model_override: ModelOverride | None,
    ) -> ModelOverride | None:
        del session, agent
        if not _prompt_has_binary_media(prompt):
            return model_override
        return _configured_media_model_name() or model_override


hooks = Hooks[None]()


@hooks.on.before_model_request
async def observe_before_model_request(
    ctx: RunContext[None],
    request_context: ModelRequestContext,
) -> ModelRequestContext:
    del ctx
    return request_context


@hooks.on.after_model_request
async def observe_after_model_request(
    ctx: RunContext[None],
    *,
    request_context: ModelRequestContext,
    response: ModelResponse,
) -> ModelResponse:
    del ctx, request_context
    return response


@hooks.on.before_tool_execute(tools=[_READ_TOOL])
async def observe_read_tool(
    ctx: RunContext[None],
    *,
    call: ToolCallPart,
    tool_def: ToolDefinition,
    args: ValidatedToolArgs,
) -> ValidatedToolArgs:
    del ctx, call, tool_def
    return args


@hooks.on.before_tool_execute(tools=[_WRITE_TOOL])
async def observe_write_tool(
    ctx: RunContext[None],
    *,
    call: ToolCallPart,
    tool_def: ToolDefinition,
    args: ValidatedToolArgs,
) -> ValidatedToolArgs:
    del ctx, call, tool_def
    return args


@hooks.on.after_tool_execute(tools=[_WRITE_TOOL])
async def observe_write_result(
    ctx: RunContext[None],
    *,
    call: ToolCallPart,
    tool_def: ToolDefinition,
    args: ValidatedToolArgs,
    result: object,
) -> object:
    del ctx, call, tool_def, args
    return result


agent = Agent(
    _default_model_name(),
    name="travel-agent",
    deps_type=type(None),
    capabilities=[hooks],
    output_type=[str, DeferredToolRequests],
    system_prompt=(
        "You are the ACP Kit travel example. "
        "Use the trip-file tools for grounded answers. "
        "When image or audio prompt blocks appear, the host may swap the active model override."
    ),
)


@agent.tool_plain
def describe_travel_surface() -> str:
    """Summarize the ACP-facing surfaces available in this travel example."""
    return "\n".join(
        (
            "Travel example features:",
            "- existing Hooks capability introspection rendered through HookProjectionMap",
            "- approval-gated file writes with ACP diffs",
            "- prompt-model override provider for image and audio prompts",
        ),
    )


@agent.tool_plain
def list_trip_files() -> str:
    """List the demo travel files available in the local workspace."""
    _ensure_travel_workspace()
    return "\n".join(sorted(path.name for path in _TRAVEL_ROOT.iterdir() if path.is_file()))


@agent.tool_plain(name=_READ_TOOL)
def read_trip_file(path: str, max_chars: int = 4000) -> str:
    """Read a travel file and return a bounded preview."""
    if max_chars <= 0:
        raise ValueError("max_chars must be positive.")
    _ensure_travel_workspace()
    file_path = _resolve_trip_path(Path(path).name if Path(path).is_absolute() else path)
    if not file_path.exists():
        raise ValueError(f"File not found: {file_path.name}")
    return _truncate_text(file_path.read_text(encoding="utf-8"), limit=max_chars)


@agent.tool_plain(name=_WRITE_TOOL, requires_approval=True)
def write_trip_file(path: str, content: str) -> str:
    """Write a travel file inside the local demo workspace."""
    _ensure_travel_workspace()
    file_path = _resolve_trip_path(Path(path).name if Path(path).is_absolute() else path)
    file_path.write_text(content, encoding="utf-8")
    return f"Wrote `{file_path.name}`."


config = AdapterConfig(
    session_store=FileSessionStore(_SESSION_STORE_ROOT),
    prompt_model_override_provider=TravelPromptModelProvider(),
    hook_projection_map=HookProjectionMap(
        hidden_event_ids=frozenset({"after_model_request"}),
        event_labels={
            "before_model_request": "Before Model",
            "before_tool_execute": "Before Execute",
            "after_tool_execute": "After Execute",
        },
    ),
    projection_maps=[
        FileSystemProjectionMap(
            default_read_tool=_READ_TOOL,
            default_write_tool=_WRITE_TOOL,
        ),
    ],
)
acp_agent = create_acp_agent(agent=agent, config=config)


def main() -> None:
    _ensure_travel_workspace()
    run_acp(agent=agent, config=config)


if __name__ == "__main__":  # pragma: no cover
    main()



from __future__ import annotations as _annotations

from collections.abc import Sequence
from dataclasses import dataclass

from acp.schema import McpCapabilities, ToolKind
from pydantic_ai.settings import ModelSettings, merge_model_settings

from ..agent_types import RuntimeAgent
from ..bridges import CapabilityBridge
from ..projection import ToolClassifier
from ..providers import ConfigOption, ModeState
from ..session.state import AcpSessionContext, JsonValue, SessionTranscriptUpdate

__all__ = ("BridgeManager",)


@dataclass(slots=True, kw_only=True)
class BridgeManager:
    base_classifier: ToolClassifier
    bridges: Sequence[CapabilityBridge]

    @property
    def tool_classifier(self) -> ToolClassifier:
        return _BridgeAwareToolClassifier(
            base_classifier=self.base_classifier,
            bridges=self.bridges,
        )

    def drain_updates(
        self,
        session: AcpSessionContext,
        agent: RuntimeAgent,
    ) -> list[SessionTranscriptUpdate]:
        updates: list[SessionTranscriptUpdate] = []
        for bridge in self.bridges:
            bridge_updates = bridge.drain_updates(session, agent)
            if bridge_updates:
                updates.extend(bridge_updates)
        return updates

    def get_config_options(
        self,
        session: AcpSessionContext,
        agent: RuntimeAgent,
    ) -> list[ConfigOption] | None:
        options: list[ConfigOption] = []
        for bridge in self.bridges:
            bridge_options = bridge.get_config_options(session, agent)
            if bridge_options:
                options.extend(bridge_options)
        return options or None

    def get_mcp_capabilities(self, agent: RuntimeAgent | None = None) -> McpCapabilities:
        merged = McpCapabilities()
        for bridge in self.bridges:
            capabilities = bridge.get_mcp_capabilities(agent)
            if capabilities is None:
                continue
            merged.http = bool(merged.http or capabilities.http)
            merged.sse = bool(merged.sse or capabilities.sse)
        return merged

    def get_mode_state(
        self,
        session: AcpSessionContext,
        agent: RuntimeAgent,
    ) -> ModeState | None:
        for bridge in self.bridges:
            mode_state = bridge.get_mode_state(session, agent)
            if mode_state is not None:
                return mode_state
        return None

    def get_model_settings(
        self,
        session: AcpSessionContext,
        agent: RuntimeAgent,
    ) -> ModelSettings | None:
        merged: ModelSettings | None = None
        for bridge in self.bridges:
            bridge_settings = bridge.get_model_settings(session, agent)
            if bridge_settings is not None:
                merged = merge_model_settings(merged, bridge_settings)
        return merged

    def get_metadata_sections(
        self,
        session: AcpSessionContext,
        agent: RuntimeAgent,
    ) -> dict[str, JsonValue]:
        metadata: dict[str, JsonValue] = {}
        for bridge in self.bridges:
            if bridge.metadata_key is None:
                continue
            bridge_metadata = bridge.get_session_metadata(session, agent)
            if bridge_metadata is not None:
                metadata[bridge.metadata_key] = bridge_metadata
        return metadata

    def set_config_option(
        self,
        session: AcpSessionContext,
        agent: RuntimeAgent,
        config_id: str,
        value: str | bool,
    ) -> list[ConfigOption] | None:
        for bridge in self.bridges:
            options = bridge.set_config_option(session, agent, config_id, value)
            if options is not None:
                return options
        return None

    def set_mode(
        self,
        session: AcpSessionContext,
        agent: RuntimeAgent,
        mode_id: str,
    ) -> ModeState | None:
        for bridge in self.bridges:
            mode_state = bridge.set_mode(session, agent, mode_id)
            if mode_state is not None:
                return mode_state
        return None


@dataclass(slots=True, frozen=True, kw_only=True)
class _BridgeAwareToolClassifier:
    base_classifier: ToolClassifier
    bridges: Sequence[CapabilityBridge]

    def classify(self, tool_name: str, raw_input: JsonValue | None = None) -> ToolKind:
        for bridge in self.bridges:
            bridge_kind = bridge.get_tool_kind(tool_name, raw_input)
            if bridge_kind is not None:
                return bridge_kind
        return self.base_classifier.classify(tool_name, raw_input)

    def approval_policy_key(self, tool_name: str, raw_input: JsonValue | None = None) -> str:
        for bridge in self.bridges:
            approval_policy_key = bridge.get_approval_policy_key(tool_name, raw_input)
            if approval_policy_key is not None:
                return approval_policy_key
        return self.base_classifier.approval_policy_key(tool_name, raw_input)



from __future__ import annotations as _annotations

from datetime import UTC, datetime
from pathlib import Path
from typing import Any

from acp.helpers import text_block
from acp.interfaces import Agent as AcpAgent
from acp.schema import (
    AgentMessageChunk,
    AgentPlanUpdate,
    AgentThoughtChunk,
    AllowedOutcome,
    AvailableCommandsUpdate,
    ConfigOptionUpdate,
    ContentToolCallContent,
    CreateTerminalResponse,
    CurrentModeUpdate,
    DeniedOutcome,
    EnvVariable,
    FileEditToolCallContent,
    KillTerminalResponse,
    PermissionOption,
    PlanEntry,
    ReadTextFileResponse,
    ReleaseTerminalResponse,
    RequestPermissionResponse,
    SessionConfigOptionBoolean,
    SessionConfigOptionSelect,
    SessionConfigSelectGroup,
    SessionConfigSelectOption,
    SessionInfoUpdate,
    SessionMode,
    TerminalOutputResponse,
    TerminalToolCallContent,
    ToolCallProgress,
    ToolCallStart,
    ToolCallUpdate,
    UsageUpdate,
    UserMessageChunk,
    WaitForTerminalExitResponse,
    WriteTextFileResponse,
)
from pydantic_acp import (
    AcpSessionContext,
    AdapterConfig,
    AdapterPromptCapabilities,
    AgentBridgeBuilder,
    AgentFactory,
    AgentSource,
    AnthropicCompactionBridge,
    ApprovalPolicy,
    ApprovalPolicyStore,
    BuiltinToolProjectionMap,
    ClientFilesystemBackend,
    ClientHostContext,
    ClientTerminalBackend,
    CompositeProjectionMap,
    ConfigOption,
    DefaultPermissionToolCallBuilder,
    ExternalHookEventBridge,
    FactoryAgentSource,
    FileSessionStore,
    FilesystemBackend,
    FileSystemProjectionMap,
    HistoryProcessorBridge,
    HookBridge,
    HookEvent,
    HookProjectionMap,
    ImageGenerationBridge,
    IncludeToolReturnSchemasBridge,
    McpBridge,
    McpCapabilityBridge,
    McpServerDefinition,
    McpToolDefinition,
    MemorySessionStore,
    ModelSelectionState,
    ModeState,
    NativeApprovalBridge,
    OpenAICompactionBridge,
    PermissionOptionSet,
    PermissionRequestContext,
    PermissionToolCallBuilder,
    PrefixToolsBridge,
    PrepareOutputToolsBridge,
    PrepareOutputToolsMode,
    PrepareToolsBridge,
    PrepareToolsMode,
    ProjectionAwareToolClassifier,
    SessionMcpBridge,
    SessionMetadataApprovalPolicyStore,
    SetToolMetadataBridge,
    SlashCommandProvider,
    SlashCommandRequest,
    SlashCommandResult,
    StaticAgentSource,
    StaticSlashCommand,
    StaticSlashCommandProvider,
    TerminalBackend,
    ThinkingBridge,
    ThreadExecutorBridge,
    ToolsetBridge,
    WebFetchBridge,
    WebSearchBridge,
    create_acp_agent,
)
from pydantic_acp.models import AdapterModel
from pydantic_acp.projection import WebToolProjectionMap
from pydantic_acp.session.state import JsonValue
from pydantic_ai import Agent
from pydantic_ai.exceptions import ApprovalRequired
from pydantic_ai.messages import ModelMessage
from pydantic_ai.models.test import TestModel
from pydantic_ai.tools import RunContext, ToolDefinition

__all__ = (
    "UTC",
    "AcpSessionContext",
    "AdapterConfig",
    "AdapterModel",
    "AdapterPromptCapabilities",
    "Agent",
    "AgentBridgeBuilder",
    "AgentFactory",
    "AgentMessageChunk",
    "AgentPlanUpdate",
    "AgentSource",
    "AllowedOutcome",
    "AnthropicCompactionBridge",
    "ApprovalPolicy",
    "ApprovalPolicyStore",
    "ApprovalRequired",
    "AsyncDemoConfigOptionsProvider",
    "AsyncDemoModesProvider",
    "AsyncDemoPlanProvider",
    "AvailableCommandsUpdate",
    "BuiltinToolProjectionMap",
    "ClientFilesystemBackend",
    "ClientHostContext",
    "ClientTerminalBackend",
    "CompositeProjectionMap",
    "ConfigOption",
    "ConfigOptionUpdate",
    "ContentToolCallContent",
    "CreateTerminalResponse",
    "CurrentModeUpdate",
    "DefaultPermissionToolCallBuilder",
    "DemoApprovalStateProvider",
    "DemoConfigOptionsProvider",
    "DemoModelsProvider",
    "DemoModesProvider",
    "DemoPlanProvider",
    "DeniedOutcome",
    "EnvVariable",
    "ExternalHookEventBridge",
    "FactoryAgentSource",
    "FileEditToolCallContent",
    "FileSessionStore",
    "FileSystemProjectionMap",
    "FilesystemBackend",
    "FilesystemRecordingClient",
    "FreeformModelsProvider",
    "HistoryProcessorBridge",
    "HookBridge",
    "HookEvent",
    "HookProjectionMap",
    "HostRecordingClient",
    "ImageGenerationBridge",
    "IncludeToolReturnSchemasBridge",
    "JsonValue",
    "KillTerminalResponse",
    "McpBridge",
    "McpCapabilityBridge",
    "McpServerDefinition",
    "McpToolDefinition",
    "MemorySessionStore",
    "ModeState",
    "ModelMessage",
    "ModelSelectionState",
    "NativeApprovalBridge",
    "OpenAICompactionBridge",
    "Path",
    "PermissionOption",
    "PermissionOptionSet",
    "PermissionRequestContext",
    "PermissionToolCallBuilder",
    "PlanEntry",
    "PrefixToolsBridge",
    "PrepareOutputToolsBridge",
    "PrepareOutputToolsMode",
    "PrepareToolsBridge",
    "PrepareToolsMode",
    "ProjectionAwareToolClassifier",
    "ReadTextFileResponse",
    "RecordingClient",
    "ReleaseTerminalResponse",
    "RequestPermissionResponse",
    "ReservedModelConfigProvider",
    "RunContext",
    "SessionConfigOptionBoolean",
    "SessionConfigOptionSelect",
    "SessionConfigSelectGroup",
    "SessionConfigSelectOption",
    "SessionInfoUpdate",
    "SessionMetadataApprovalPolicyStore",
    "SessionMode",
    "SetToolMetadataBridge",
    "SessionMcpBridge",
    "SlashCommandProvider",
    "SlashCommandRequest",
    "SlashCommandResult",
    "StaticAgentSource",
    "StaticSlashCommand",
    "StaticSlashCommandProvider",
    "TerminalBackend",
    "TerminalOutputResponse",
    "TerminalRecordingClient",
    "TerminalToolCallContent",
    "TestModel",
    "ThinkingBridge",
    "ThreadExecutorBridge",
    "ToolCallProgress",
    "ToolCallStart",
    "ToolCallUpdate",
    "ToolDefinition",
    "ToolsetBridge",
    "UsageUpdate",
    "UserMessageChunk",
    "WaitForTerminalExitResponse",
    "WebFetchBridge",
    "WebSearchBridge",
    "WebToolProjectionMap",
    "WriteTextFileResponse",
    "agent_message_texts",
    "create_acp_agent",
    "datetime",
    "text_block",
)


class RecordingClient:
    def __init__(self) -> None:
        self.updates: list[tuple[str, Any]] = []
        self.permission_option_ids: list[tuple[str, list[str], ToolCallUpdate]] = []
        self.permission_option_names: list[tuple[str, list[str], ToolCallUpdate]] = []
        self.permission_responses: list[RequestPermissionResponse] = []

    def queue_permission_selected(self, option_id: str) -> None:
        self.permission_responses.append(
            RequestPermissionResponse(
                outcome=AllowedOutcome(outcome="selected", option_id=option_id),
            ),
        )

    def queue_permission_cancelled(self) -> None:
        self.permission_responses.append(
            RequestPermissionResponse(outcome=DeniedOutcome(outcome="cancelled")),
        )

    async def request_permission(
        self,
        options: list[PermissionOption],
        session_id: str,
        tool_call: ToolCallUpdate,
        **kwargs: Any,
    ) -> RequestPermissionResponse:
        del kwargs
        self.permission_option_ids.append(
            (session_id, [option.option_id for option in options], tool_call),
        )
        self.permission_option_names.append(
            (session_id, [option.name for option in options], tool_call),
        )
        if not self.permission_responses:
            raise AssertionError("unexpected permission request")
        return self.permission_responses.pop(0)

    async def session_update(
        self,
        session_id: str,
        update: (
            UserMessageChunk
            | AgentMessageChunk
            | AgentThoughtChunk
            | ToolCallStart
            | ToolCallProgress
            | AgentPlanUpdate
            | AvailableCommandsUpdate
            | CurrentModeUpdate
            | ConfigOptionUpdate
            | SessionInfoUpdate
            | UsageUpdate
        ),
        **kwargs: Any,
    ) -> None:
        del kwargs
        self.updates.append((session_id, update))

    async def write_text_file(
        self,
        content: str,
        path: str,
        session_id: str,
        **kwargs: Any,
    ) -> WriteTextFileResponse | None:
        del content, path, session_id, kwargs
        raise AssertionError("filesystem flow is not part of this test")

    async def read_text_file(
        self,
        path: str,
        session_id: str,
        limit: int | None = None,
        line: int | None = None,
        **kwargs: Any,
    ) -> ReadTextFileResponse:
        del path, session_id, limit, line, kwargs
        raise AssertionError("filesystem flow is not part of this test")

    async def create_terminal(
        self,
        command: str,
        session_id: str,
        args: list[str] | None = None,
        cwd: str | None = None,
        env: list[EnvVariable] | None = None,
        output_byte_limit: int | None = None,
        **kwargs: Any,
    ) -> CreateTerminalResponse:
        del command, session_id, args, cwd, env, output_byte_limit, kwargs
        raise AssertionError("terminal flow is not part of this test")

    async def terminal_output(
        self,
        session_id: str,
        terminal_id: str,
        **kwargs: Any,
    ) -> TerminalOutputResponse:
        del session_id, terminal_id, kwargs
        raise AssertionError("terminal flow is not part of this test")

    async def release_terminal(
        self,
        session_id: str,
        terminal_id: str,
        **kwargs: Any,
    ) -> ReleaseTerminalResponse | None:
        del session_id, terminal_id, kwargs
        raise AssertionError("terminal flow is not part of this test")

    async def wait_for_terminal_exit(
        self,
        session_id: str,
        terminal_id: str,
        **kwargs: Any,
    ) -> WaitForTerminalExitResponse:
        del session_id, terminal_id, kwargs
        raise AssertionError("terminal flow is not part of this test")

    async def kill_terminal(
        self,
        session_id: str,
        terminal_id: str,
        **kwargs: Any,
    ) -> KillTerminalResponse | None:
        del session_id, terminal_id, kwargs
        raise AssertionError("terminal flow is not part of this test")

    async def ext_method(self, method: str, params: dict[str, Any]) -> dict[str, Any]:
        raise AssertionError(f"unexpected extension method: {method!r} {params!r}")

    async def ext_notification(self, method: str, params: dict[str, Any]) -> None:
        raise AssertionError(f"unexpected extension notification: {method!r} {params!r}")

    def on_connect(self, conn: AcpAgent) -> None:  # type: ignore[misc]
        del conn


def agent_message_texts(client: RecordingClient) -> list[str]:
    messages: list[str] = []
    current_message_id: str | None = None
    current_text = ""
    anonymous_message_count = 0

    for _, update in client.updates:
        if not isinstance(update, AgentMessageChunk):
            continue
        message_id = update.message_id
        if message_id is None:
            message_id = f"anonymous:{anonymous_message_count}"
            anonymous_message_count += 1
        if current_message_id != message_id:
            if current_message_id is not None:
                messages.append(current_text)
            current_message_id = message_id
            current_text = ""
        current_text += update.content.text

    if current_message_id is not None:
        messages.append(current_text)
    return messages


class FilesystemRecordingClient(RecordingClient):
    def __init__(self) -> None:
        super().__init__()
        self.read_calls: list[tuple[str, str, int | None, int | None]] = []
        self.write_calls: list[tuple[str, str, str]] = []
        self.write_response: WriteTextFileResponse | None = WriteTextFileResponse()

    async def write_text_file(
        self,
        content: str,
        path: str,
        session_id: str,
        **kwargs: Any,
    ) -> WriteTextFileResponse | None:
        del kwargs
        self.write_calls.append((session_id, path, content))
        return self.write_response

    async def read_text_file(
        self,
        path: str,
        session_id: str,
        limit: int | None = None,
        line: int | None = None,
        **kwargs: Any,
    ) -> ReadTextFileResponse:
        del kwargs
        self.read_calls.append((session_id, path, limit, line))
        return ReadTextFileResponse(content=f"file:{path}:{line}:{limit}")


class TerminalRecordingClient(RecordingClient):
    def __init__(self) -> None:
        super().__init__()
        self.create_calls: list[
            tuple[
                str,
                str,
                list[str] | None,
                str | None,
                list[EnvVariable] | None,
                int | None,
            ]
        ] = []
        self.output_calls: list[tuple[str, str]] = []
        self.release_calls: list[tuple[str, str]] = []
        self.wait_calls: list[tuple[str, str]] = []
        self.kill_calls: list[tuple[str, str]] = []
        self.release_response: ReleaseTerminalResponse | None = ReleaseTerminalResponse()
        self.kill_response: KillTerminalResponse | None = KillTerminalResponse()
        self.wait_response = WaitForTerminalExitResponse(exit_code=0)

    async def create_terminal(
        self,
        command: str,
        session_id: str,
        args: list[str] | None = None,
        cwd: str | None = None,
        env: list[EnvVariable] | None = None,
        output_byte_limit: int | None = None,
        **kwargs: Any,
    ) -> CreateTerminalResponse:
        del kwargs
        self.create_calls.append((session_id, command, args, cwd, env, output_byte_limit))
        return CreateTerminalResponse(terminal_id="terminal-1")

    async def terminal_output(
        self,
        session_id: str,
        terminal_id: str,
        **kwargs: Any,
    ) -> TerminalOutputResponse:
        del kwargs
        self.output_calls.append((session_id, terminal_id))
        return TerminalOutputResponse(output="terminal-output", truncated=False)

    async def release_terminal(
        self,
        session_id: str,
        terminal_id: str,
        **kwargs: Any,
    ) -> ReleaseTerminalResponse | None:
        del kwargs
        self.release_calls.append((session_id, terminal_id))
        return self.release_response

    async def wait_for_terminal_exit(
        self,
        session_id: str,
        terminal_id: str,
        **kwargs: Any,
    ) -> WaitForTerminalExitResponse:
        del kwargs
        self.wait_calls.append((session_id, terminal_id))
        return self.wait_response

    async def kill_terminal(
        self,
        session_id: str,
        terminal_id: str,
        **kwargs: Any,
    ) -> KillTerminalResponse | None:
        del kwargs
        self.kill_calls.append((session_id, terminal_id))
        return self.kill_response


class HostRecordingClient(FilesystemRecordingClient, TerminalRecordingClient):
    pass


class DemoModelsProvider:
    def __init__(self) -> None:
        self._models = [
            AdapterModel(
                model_id="provider-model-a",
                name="Provider Model A",
                override=TestModel(
                    custom_output_text="provider:model-a",
                    model_name="provider-model-a",
                ),
            ),
            AdapterModel(
                model_id="provider-model-b",
                name="Provider Model B",
                override=TestModel(
                    custom_output_text="provider:model-b",
                    model_name="provider-model-b",
                ),
            ),
        ]
        self._model_ids = {model.model_id for model in self._models}

    async def get_model_state(
        self,
        session: AcpSessionContext,
        agent: Agent[Any, Any],
    ) -> ModelSelectionState:
        del agent
        current_model_id = str(session.config_values.get("provider_model", "provider-model-a"))
        return ModelSelectionState(
            available_models=list(self._models),
            current_model_id=current_model_id,
            config_option_name="Provider Model",
            config_option_description="Provider-backed model override.",
        )

    async def set_model(
        self,
        session: AcpSessionContext,
        agent: Agent[Any, Any],
        model_id: str,
    ) -> ModelSelectionState:
        if model_id not in self._model_ids:
            raise AssertionError(f"unexpected model id: {model_id}")
        session.config_values["provider_model"] = model_id
        return await self.get_model_state(session, agent)


class DemoModesProvider:
    def __init__(self) -> None:
        self._modes = [
            SessionMode(
                id="chat",
                name="Chat",
                description="General conversational mode.",
            ),
            SessionMode(
                id="review",
                name="Review",
                description="Focused review mode.",
            ),
        ]
        self._mode_ids = {mode.id for mode in self._modes}

    def get_mode_state(
        self,
        session: AcpSessionContext,
        agent: Agent[Any, Any],
    ) -> ModeState:
        del agent
        current_mode_id = str(session.config_values.get("provider_mode", "chat"))
        return ModeState(
            modes=list(self._modes),
            current_mode_id=current_mode_id,
        )

    def set_mode(
        self,
        session: AcpSessionContext,
        agent: Agent[Any, Any],
        mode_id: str,
    ) -> ModeState:
        if mode_id not in self._mode_ids:
            raise AssertionError(f"unexpected mode id: {mode_id}")
        session.config_values["provider_mode"] = mode_id
        return self.get_mode_state(session, agent)


class DemoConfigOptionsProvider:
    def get_config_options(
        self,
        session: AcpSessionContext,
        agent: Agent[Any, Any],
    ) -> list[ConfigOption]:
        del agent
        stream_enabled = bool(session.config_values.get("stream_enabled", False))
        return [
            SessionConfigOptionBoolean(
                id="stream_enabled",
                name="Streaming",
                category="runtime",
                description="Enable streamed responses.",
                type="boolean",
                current_value=stream_enabled,
            ),
        ]

    def set_config_option(
        self,
        session: AcpSessionContext,
        agent: Agent[Any, Any],
        config_id: str,
        value: str | bool,
    ) -> list[ConfigOption] | None:
        if config_id != "stream_enabled" or not isinstance(value, bool):
            return None
        session.config_values["stream_enabled"] = value
        return self.get_config_options(session, agent)


class DemoPlanProvider:
    def get_plan(
        self,
        session: AcpSessionContext,
        agent: Agent[Any, Any],
    ) -> list[PlanEntry]:
        del agent
        current_mode = str(session.config_values.get("mode", "chat"))
        stream_enabled = bool(session.config_values.get("stream_enabled", False))
        return [
            PlanEntry(content=f"mode:{current_mode}", priority="high", status="in_progress"),
            PlanEntry(
                content=f"stream:{str(stream_enabled).lower()}",
                priority="low",
                status="pending",
            ),
        ]


class FreeformModelsProvider:
    async def get_model_state(
        self,
        session: AcpSessionContext,
        agent: Agent[Any, Any],
    ) -> ModelSelectionState:
        del agent
        current_model_id = str(session.config_values.get("freeform_model", "custom-model-a"))
        return ModelSelectionState(
            available_models=[
                AdapterModel(
                    model_id="custom-model-a",
                    name="Custom Model A",
                    override="custom-model-a",
                ),
            ],
            current_model_id=current_model_id,
            allow_any_model_id=True,
        )

    async def set_model(
        self,
        session: AcpSessionContext,
        agent: Agent[Any, Any],
        model_id: str,
    ) -> ModelSelectionState:
        session.config_values["freeform_model"] = model_id
        return await self.get_model_state(session, agent)


class ReservedModelConfigProvider:
    def get_config_options(
        self,
        session: AcpSessionContext,
        agent: Agent[Any, Any],
    ) -> list[ConfigOption]:
        del agent
        current_value = str(session.config_values.get("provider_model_option", "provider-model-a"))
        return [
            SessionConfigOptionSelect(
                id="model",
                name="Provider Model Config",
                category="runtime",
                description="Provider-owned model config option.",
                type="select",
                current_value=current_value,
                options=[
                    SessionConfigSelectOption(value="provider-model-a", name="Provider Model A"),
                    SessionConfigSelectOption(value="provider-model-b", name="Provider Model B"),
                ],
            ),
        ]

    def set_config_option(
        self,
        session: AcpSessionContext,
        agent: Agent[Any, Any],
        config_id: str,
        value: str | bool,
    ) -> list[ConfigOption] | None:
        if config_id != "model" or not isinstance(value, str):
            return None
        session.config_values["provider_model_option"] = value
        return self.get_config_options(session, agent)


class AsyncDemoModesProvider:
    def __init__(self) -> None:
        self._delegate = DemoModesProvider()

    async def get_mode_state(
        self,
        session: AcpSessionContext,
        agent: Agent[Any, Any],
    ) -> ModeState:
        return self._delegate.get_mode_state(session, agent)

    async def set_mode(
        self,
        session: AcpSessionContext,
        agent: Agent[Any, Any],
        mode_id: str,
    ) -> ModeState:
        return self._delegate.set_mode(session, agent, mode_id)


class AsyncDemoConfigOptionsProvider:
    def __init__(self) -> None:
        self._delegate = DemoConfigOptionsProvider()

    async def get_config_options(
        self,
        session: AcpSessionContext,
        agent: Agent[Any, Any],
    ) -> list[ConfigOption]:
        return self._delegate.get_config_options(session, agent)

    async def set_config_option(
        self,
        session: AcpSessionContext,
        agent: Agent[Any, Any],
        config_id: str,
        value: str | bool,
    ) -> list[ConfigOption] | None:
        return self._delegate.set_config_option(session, agent, config_id, value)


class AsyncDemoPlanProvider:
    def __init__(self) -> None:
        self._delegate = DemoPlanProvider()

    async def get_plan(
        self,
        session: AcpSessionContext,
        agent: Agent[Any, Any],
    ) -> list[PlanEntry]:
        return self._delegate.get_plan(session, agent)


class DemoApprovalStateProvider:
    def get_approval_state(
        self,
        session: AcpSessionContext,
        agent: Agent[Any, Any],
    ) -> dict[str, JsonValue]:
        del agent
        return {
            "policy": "session",
            "remembered": bool(session.config_values.get("remembered_approval", True)),
        }



from __future__ import annotations as _annotations

import asyncio
import importlib.util
from pathlib import Path
from typing import Any, cast

from pydantic_ai import (
    ModelRequest,
    ModelResponse,
    TextPart,
    ToolCallPart,
    ToolReturnPart,
)
from pydantic_ai.messages import UserPromptPart
from pydantic_ai.models.function import AgentInfo, FunctionModel

from .pydantic.support import (
    FileEditToolCallContent,
    RecordingClient,
    ToolCallProgress,
    ToolCallStart,
    create_acp_agent,
    text_block,
)


def _load_demo_module():
    module_path = Path(__file__).resolve().parents[1] / "examples" / "pydantic" / "travel_agent.py"
    spec = importlib.util.spec_from_file_location("travel_agent_demo", module_path)
    assert spec is not None
    assert spec.loader is not None
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def _latest_user_prompt(messages: list[ModelRequest | ModelResponse]) -> str:
    for message in reversed(messages):
        if not isinstance(message, ModelRequest):
            continue
        for part in reversed(message.parts):
            if isinstance(part, UserPromptPart):  # pragma: no branch
                content = part.content
                if isinstance(content, str):
                    return content
    return ""


def _travel_demo_model(
    messages: list[ModelRequest | ModelResponse],
    info: AgentInfo,
) -> ModelResponse:
    del info
    latest_message = messages[-1]
    if isinstance(latest_message, ModelRequest):  # pragma: no branch
        for part in latest_message.parts:
            if isinstance(part, ToolReturnPart):
                return ModelResponse(parts=[TextPart(f"{part.tool_name}: {part.content}")])

    prompt = _latest_user_prompt(messages)
    if prompt == "read trip file itinerary.md":
        return ModelResponse(
            parts=[ToolCallPart("read_trip_file", {"path": "itinerary.md", "max_chars": 4000})],
        )
    if prompt == "write trip file scratch.txt: hello from the native demo":
        return ModelResponse(
            parts=[
                ToolCallPart(
                    "write_trip_file",
                    {"path": "scratch.txt", "content": "hello from the native demo"},
                ),
            ],
        )
    return ModelResponse(parts=[TextPart("Travel demo mode is active.")])


demo = _load_demo_module()


def test_native_pydantic_agent_helpers_cover_text_and_fallback_paths() -> None:
    empty_prompt = _latest_user_prompt([])
    structured_prompt = _latest_user_prompt(
        [
            ModelRequest(parts=[UserPromptPart(content=["not-a-string"])]),
            ModelResponse(parts=[TextPart("ignored")]),
        ],
    )
    plain_prompt = _latest_user_prompt([ModelRequest(parts=[UserPromptPart(content="hello")])])

    assert empty_prompt == ""
    assert structured_prompt == ""
    assert plain_prompt == "hello"

    tool_return_response = _travel_demo_model(
        [ModelRequest(parts=[ToolReturnPart(tool_name="read_trip_file", content="done")])],
        cast("Any", object()),
    )
    fallback_response = _travel_demo_model(
        [ModelRequest(parts=[UserPromptPart(content="something else")])],
        cast("Any", object()),
    )
    response_passthrough = _travel_demo_model(
        [ModelResponse(parts=[TextPart("already done")])],
        cast("Any", object()),
    )

    assert tool_return_response.parts == [TextPart("read_trip_file: done")]
    assert fallback_response.parts == [TextPart("Travel demo mode is active.")]
    assert response_passthrough.parts == [TextPart("Travel demo mode is active.")]


def test_native_pydantic_agent_read_prompt_emits_hook_and_diff(
    tmp_path: Path,
    monkeypatch,
) -> None:
    monkeypatch.chdir(tmp_path)
    monkeypatch.setattr(demo, "_TRAVEL_ROOT", tmp_path / "native-demo")
    demo._ensure_travel_workspace()
    monkeypatch.setattr(
        demo.agent,
        "model",
        FunctionModel(_travel_demo_model, model_name="travel-demo-model"),
    )

    adapter = create_acp_agent(agent=demo.agent, config=demo.config)
    client = RecordingClient()
    adapter.on_connect(client)

    session = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    asyncio.run(
        adapter.prompt(
            prompt=[text_block("read trip file itinerary.md")],
            session_id=session.session_id,
        ),
    )

    hook_titles = [
        update.title for _, update in client.updates if isinstance(update, ToolCallStart)
    ]
    assert "Hook Before Model (observe_before_model_request)" in hook_titles

    read_updates = [
        update
        for _, update in client.updates
        if isinstance(update, ToolCallProgress) and update.title == "read_trip_file"
    ]
    assert len(read_updates) == 1
    assert read_updates[0].content is not None
    diff = read_updates[0].content[0]
    assert isinstance(diff, FileEditToolCallContent)
    assert diff.path.endswith("itinerary.md")
    assert diff.old_text == ""
    assert "Travel Brief" in diff.new_text


def test_native_pydantic_agent_write_prompt_emits_hook_and_diff(
    tmp_path: Path,
    monkeypatch,
) -> None:
    monkeypatch.chdir(tmp_path)
    monkeypatch.setattr(demo, "_TRAVEL_ROOT", tmp_path / "native-demo")
    demo._ensure_travel_workspace()
    monkeypatch.setattr(
        demo.agent,
        "model",
        FunctionModel(_travel_demo_model, model_name="travel-demo-model"),
    )

    adapter = create_acp_agent(agent=demo.agent, config=demo.config)
    client = RecordingClient()
    client.queue_permission_selected("allow_once")
    adapter.on_connect(client)

    session = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    asyncio.run(
        adapter.prompt(
            prompt=[text_block("write trip file scratch.txt: hello from the native demo")],
            session_id=session.session_id,
        ),
    )

    write_start = next(
        cast("Any", update)
        for _, update in client.updates
        if isinstance(update, ToolCallStart) and update.title == "write_trip_file"
    )
    assert write_start.content is not None
    diff = write_start.content[0]
    assert isinstance(diff, FileEditToolCallContent)
    assert diff.path.endswith("scratch.txt")
    assert diff.old_text == ""
    assert diff.new_text == "hello from the native demo"

    hook_titles = [
        update.title for _, update in client.updates if isinstance(update, ToolCallStart)
    ]
    assert "Hook Before Execute [write_trip_file] (observe_write_tool)" in hook_titles
    assert (tmp_path / "native-demo" / "scratch.txt").read_text(encoding="utf-8") == (
        "hello from the native demo"
    )
